from PymoNNto import *
import copy
import pickle
import os
import marshal

compile_class = type(compile('1+1', '<string>', 'eval'))
base_obj_type = TaggableObjectBase#NetworkObjectBase

def add_to_compiled_save_list_and_get_save_str(obj, compiled, compiled_path):
    ct = len(compiled)+1
    s = '__compile#'+str(ct)+'.bin'
    compiled[s] = obj
    marshal.dump(obj, open(compiled_path + s, "wb"))
    return s

def clear_compiled_code(obj, visited=[], compiled={}, compiled_path=None):

    if isinstance(obj, base_obj_type) and (obj not in visited):
        visited.append(obj)
        d = obj.__dict__
        for key in d:
            if type(getattr(obj, key)) in [list, dict] or isinstance(getattr(obj, key), base_obj_type):
                clear_compiled_code(d[key], visited, compiled, compiled_path)

            if type(getattr(obj, key)) is compile_class:
                setattr(obj, key, add_to_compiled_save_list_and_get_save_str(getattr(obj, key), compiled, compiled_path))

    visited.append(obj)

    if type(obj) is list:
        for i in range(len(obj)):
            if type(obj[i]) in [list, dict] or isinstance(obj[i], base_obj_type):
                clear_compiled_code(obj[i], visited, compiled, compiled_path)

            if type(obj[i]) is compile_class:
                obj[i] = add_to_compiled_save_list_and_get_save_str(obj[i], compiled, compiled_path)

    if type(obj) is dict:
        for sub_obj_key in obj:
            if type(obj[sub_obj_key]) in [list, dict] or isinstance(obj[sub_obj_key], base_obj_type):
                clear_compiled_code(obj[sub_obj_key], visited, compiled, compiled_path)

            if type(obj[sub_obj_key]) is compile_class:
                obj[sub_obj_key] = add_to_compiled_save_list_and_get_save_str(obj[sub_obj_key], compiled, compiled_path)

    return compiled

def load_compiled_object(file, path):
    return marshal.load(open(path + file, "rb"))


def add_compiled_code(obj, visited=[], path=''):

    if isinstance(obj, base_obj_type) and (obj not in visited):
        visited.append(obj)
        d = obj.__dict__
        for key in d:
            if type(getattr(obj, key)) in [list, dict] or isinstance(getattr(obj, key), base_obj_type):
                add_compiled_code(d[key], visited, path)

            if type(getattr(obj, key)) is str and '__compile#' in getattr(obj, key):
                setattr(obj, key, load_compiled_object(getattr(obj, key), path))

    visited.append(obj)

    if type(obj) is list:
        for i in range(len(obj)):
            if type(obj[i]) in [list, dict] or isinstance(obj[i], base_obj_type):
                add_compiled_code(obj[i], visited, path)

            if type(obj[i]) is str and '__compile#' in obj[i]:
                obj[i] = load_compiled_object(obj[i], path)

    if type(obj) is dict:
        for sub_obj_key in obj:
            if type(obj[sub_obj_key]) in [list, dict] or isinstance(obj[sub_obj_key], base_obj_type):
                add_compiled_code(obj[sub_obj_key], visited, path)

            if type(obj[sub_obj_key]) is str and '__compile#' in obj[sub_obj_key]:
                obj[sub_obj_key] = load_compiled_object(obj[sub_obj_key], path)


def save_network(network, filename):
    folder = get_data_folder() + '/NetworkStates/'

    if not os.path.exists(folder):
        try:
            os.mkdir(folder)
        except:
            print('was not able to create '+folder+' folder')

    folder +=  filename + '/'

    if not os.path.exists(folder):
        try:
            os.mkdir(folder)
        except:
            print('was not able to create '+folder+' folder')

    net = copy.deepcopy(network)
    clear_compiled_code(net, compiled_path=folder)

    file = folder + 'network.netstate'
    pickle.dump(net, open(file, 'wb'))
    print('networks saved as', file)



def load_network(filename):
    path=get_data_folder() + '/NetworkStates/'+filename+'/'
    file = path + 'network.netstate'
    network = pickle.load(open(file, 'rb'))
    add_compiled_code(network, path=path)
    return network



from PymoNNto import *

class Static_Classification(AnalysisModule):

    def initialize(self, neurons):
        self.add_tag('classifier')
        self.name = self.parameter('name', super()._get_base_name_())
        classes = self.parameter('classes', None) #array with size neurons.size where each entry is a number that defines the neurons class
        self.save_result(self.name, classes) #stored in the results array (one classifier can generate multiple classifications)

    def _get_base_name_(self):
        return self.name



from PymoNNto.NetworkCore.Base_Attachable_Modules import *
import copy


class SynapseGroup(NetworkObjectBase):

    def __init__(self, net, src, dst, tag=None, behavior={}):

        if type(src) is str:
            s=src
            src = net[src, 0]
            if src is None:
                print(s, 'src not found for', tag)

        if type(dst) is str:
            d=dst
            dst = net[dst, 0]
            if dst is None:
                print(d, 'dst not found for', tag)

        if tag is None and net is not None:
            tag = 'SynapseGroup_'+str(len(net.SynapseGroups)+1)

        super().__init__(tag, net, behavior)
        self.add_tag('syn')

        if src is not None and dst is not None:
            if len(src.tags) > 0 and len(dst.tags) > 0:
                self.add_tag(src.tag+'_to_'+dst.tag)
        else:
            print('Warning: Not able to find source or taget for SynapseGroup '+str(self.tags))

        if net is not None:
            net.SynapseGroups.append(self)
            for tag in self.tags:
                if not hasattr(net, tag):
                    setattr(net, tag, self)

        self.recording = True

        self.src = src
        self.dst = dst
        self.enabled = True
        self.group_weighting = 1

        for ng in self.network.NeuronGroups:
            for tag in self.tags + ['All']:
                if tag not in ng.afferent_synapses:
                    ng.afferent_synapses[tag] = []
                if tag not in ng.efferent_synapses:
                    ng.efferent_synapses[tag] = []

        if self.dst.BaseNeuronGroup == self.dst: #only add to NeuronGroup not to NeuronSubGroup
            for tag in self.tags+['All']:
                self.dst.afferent_synapses[tag].append(self)

        if self.src.BaseNeuronGroup == self.src:
            for tag in self.tags+['All']:
                self.src.efferent_synapses[tag].append(self)





    def __str__(self):
        if self.network.transposed_synapse_matrix_mode:
            result = 'SynapseGroup' + str(self.tags) + '(S' + str(self.src.size) + 'xD' + str(self.dst.size) + '){'
        else:
            result = 'SynapseGroup'+str(self.tags)+'(D'+str(self.dst.size)+'xS'+str(self.src.size)+'){'
        for k in sorted(list(self.behavior.keys())):
            result += str(k) + ':' + str(self.behavior[k])+','
        return result+'}'

    def set_var(self, key, value):
        setattr(self,key,value)
        return self

    @property
    def def_dtype(self):
        return self.network.def_dtype

    @property
    def iteration(self):
        return self.network.iteration

    def matrix_dim(self):
        if self.network.transposed_synapse_matrix_mode:
            return self.src.size, self.dst.size
        else:
            return self.dst.size, self.src.size

    def get_random_synapse_mat_fixed(self, min_number_of_synapses=0):
        dim = self.matrix_dim()
        result = np.zeros(dim)
        if min_number_of_synapses != 0:
            for i in range(dim[0]):
                synapses = np.random.choice(list(range(dim[1])), size=int(min_number_of_synapses), replace=False)
                result[i, synapses] = np.random.rand(len(synapses))
        return result#*np.random.rand(dim)

    def matrix(self, mode='zeros()', scale=None, density=None, plot=False):
        return self._get_mat(mode=mode, dim=(self.matrix_dim()), scale=scale, density=density, plot=plot)

    matrix = matrix
    mat = matrix
    get_synapse_mat = matrix

    def get_synapse_group_size_factor(self, synapse_group, synapse_type):

        total_weighting = 0
        for s in synapse_group.dst.synapses(afferent, synapse_type):
            total_weighting += s.group_weighting

        total = 0
        for s in synapse_group.dst.synapses(afferent, synapse_type):
            total += s.src.size*s.src.group_weighting

        return total_weighting/total*synapse_group.src.size*synapse_group.group_weighting

    def get_distance_mat(self, radius, src_x=None, src_y=None, dst_x=None, dst_y=None):
        if src_x is None: src_x = self.src.x
        if src_y is None: src_y = self.src.y
        if dst_x is None: dst_x = self.dst.x
        if dst_y is None: dst_y = self.dst.y

        result_syn_mat = np.zeros((len(dst_x), len(src_x)))

        for d_n in range(len(dst_x)):
            dx = np.abs(src_x - dst_x[d_n])
            dy = np.abs(src_y - dst_y[d_n])

            dist = np.sqrt(dx * dx + dy * dy)
            inv_dist = np.clip(radius - dist, 0.0, None)
            inv_dist /= np.max(inv_dist)

            result_syn_mat[d_n] = inv_dist

        return result_syn_mat


    def get_ring_mat(self, radius, inner_exp, src_x=None, src_y=None, dst_x=None, dst_y=None):
        dm = self.get_distance_mat(radius, src_x, src_y, dst_x, dst_y)
        ring = np.clip(dm - np.power(dm, inner_exp)*1.5, 0.0, None)
        return ring/np.max(ring)

    def get_max_receptive_field_size(self):
        max_dx = 1
        max_dy = 1
        max_dz = 1

        for i in range(self.dst.size):
            if type(self.enabled) is np.ndarray:
                mask = self.enabled[i]
            else:
                mask = self.enabled

            if np.sum(mask)>0:
                x = self.dst.x[i]
                y = self.dst.y[i]
                z = self.dst.z[i]

                sx_v = self.src.x[mask]
                sy_v = self.src.y[mask]
                sz_v = self.src.z[mask]

                max_dx = np.maximum(np.max(np.abs(x-sx_v)), max_dx)
                max_dy = np.maximum(np.max(np.abs(y-sy_v)), max_dy)
                max_dz = np.maximum(np.max(np.abs(z-sz_v)), max_dz)

        return max_dx, max_dy, max_dz

    def get_sub_synapse_group(self, src_mask, dst_mask):
        result = SynapseGroup(self.src.subGroup(src_mask), self.dst.subGroup(dst_mask), net=None, behavior={})

        # partition enabled update
        if type(self.enabled) is np.ndarray:
            mat_mask = dst_mask[:, None] * src_mask[None, :]
            result.enabled = self.enabled[mat_mask].copy().reshape(result.matrix_dim())

        # copy al attributes
        sgd = self.__dict__
        for key in sgd:
            if key == 'behavior':
                for k in self.behavior:
                    result.behavior[k] = copy.copy(self.behavior[k])
            elif key not in ['src', 'dst', 'enabled', '_mat_eval_dict']:
                setattr(result, key, copy.copy(sgd[key]))

        return result

    def ignore_transpose_mode(self, W):# always returns DxS matrix
        if self.network.transposed_synapse_matrix_mode:
            return W.T
        else:
            return W




from PymoNNto.Exploration.Network_UI.TabBase import *

class hist_tab(TabBase):

    def __init__(self, weight_attr='W', title='Weight Dist.', timesteps=1000, mask_param='Input_Mask', mask_color_add=(-100, -100, -100)):#mask_param=None #
        super().__init__(title)
        self.weight_attr = weight_attr
        self.timesteps = timesteps

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        self.additionaltab = Network_UI.add_tab(title=self.title)

        self.weight_hist_plots = {}
        self.net_weight_hist_plots = {}

        for i,transmitter in enumerate(Network_UI.transmitters):
            if i>0:
                Network_UI.tab.add_row()

            self.weight_hist_plots[transmitter] = Network_UI.tab.add_plot(title=transmitter + ' selected weight hist', x_label=transmitter + ' synapse size', y_label='Frequency')
            self.net_weight_hist_plots[transmitter] = Network_UI.tab.add_plot(title=transmitter + ' network weight hist', x_label=transmitter + ' synapse size', y_label='Frequency')

        Network_UI.tab.add_row()
        #self.min_hist_slider = QSlider(1)  # QtCore.Horizontal
        #self.min_hist_slider.setMinimum(-1)
        #self.min_hist_slider.setMaximum(10)
        #self.min_hist_slider.setSliderPosition(0)
        #self.min_hist_slider.mouseReleaseEvent = Network_UI.static_update_func
        #self.min_hist_slider.setToolTip('slide to cut away smallest weights')
        #Network_UI.tab.add_widget(self.min_hist_slider)  # , stretch=0.1

        Network_UI.tab.add_widget(QLabel('min: '))
        self.qsb_min = QDoubleSpinBox()
        self.qsb_min.setDecimals(5)
        self.qsb_min.setValue(0.0)
        self.qsb_min.setSingleStep(0.00001)
        Network_UI.tab.add_widget(self.qsb_min)

        Network_UI.tab.add_widget(QLabel('max: '))
        self.qsb_max = QDoubleSpinBox()
        self.qsb_max.setDecimals(5)
        self.qsb_max.setValue(1.0)
        self.qsb_max.setSingleStep(0.00001)
        Network_UI.tab.add_widget(self.qsb_max)

        #Network_UI.tab.add_row()
        Network_UI.tab.add_widget(QLabel('bins: '))
        self.bin_slider = QSlider(1)  # QtCore.Horizontal
        self.bin_slider.setMinimum(1)
        self.bin_slider.setMaximum(100)
        self.bin_slider.setSliderPosition(50)
        self.bin_slider.mouseReleaseEvent = Network_UI.static_update_func
        self.bin_slider.setToolTip('slide to change bin count')
        Network_UI.tab.add_widget(self.bin_slider)  # , stretch=0.1

    def update_Synapse_Historgrams(self, Network_UI, group, net_color_input):
        #msl = self.min_hist_slider.sliderPosition() * 0.001

        min_weight = self.qsb_min.value()
        max_weight = self.qsb_max.value()

        bins = self.bin_slider.sliderPosition()

        for transmitter in Network_UI.transmitters:

            self.net_weight_hist_plots[transmitter].clear()
            self.weight_hist_plots[transmitter].clear()

            glu_syns = group.synapses(afferent, transmitter)
            if len(glu_syns) > 0:

                GLU_syn_list = get_combined_syn_mats(glu_syns, None, self.weight_attr)
                GLU_syn_list_en = get_combined_syn_mats(glu_syns, None, "enabled")
                if len(GLU_syn_list) > 0:
                    GLU_syn = GLU_syn_list[list(GLU_syn_list.keys())[0]]
                    en_mask = GLU_syn_list_en[list(GLU_syn_list_en.keys())[0]].astype(bool)*(GLU_syn > min_weight)*(GLU_syn < max_weight)#GLU_syn > msl

                    self.net_weight_hist_plots[transmitter].clear()
                    y, x = np.histogram(GLU_syn[en_mask], bins=bins)#[GLU_syn[not_input_mask] > msl]
                    curve = pg.PlotCurveItem(x, y, stepMode=True, fillLevel=0, brush=group.color)
                    self.net_weight_hist_plots[transmitter].addItem(curve)

                    self.weight_hist_plots[transmitter].clear()
                    y, x = np.histogram(GLU_syn[Network_UI.selected_neuron_mask()][en_mask[Network_UI.selected_neuron_mask()]], bins=bins)#[selected_neuron_GLU_syn > msl]
                    curve = pg.PlotCurveItem(x, y, stepMode=True, fillLevel=0, brush=Network_UI.neuron_select_color)
                    self.weight_hist_plots[transmitter].addItem(curve)


    def update(self, Network_UI):
        if self.additionaltab.isVisible():

            group = Network_UI.selected_neuron_group()
            n=group#for eval comand

            mca = (0,0,0)

            net_color_input = np.clip([group.color[0] + mca[0], group.color[1] + mca[1], group.color[2] + mca[2], 255], 0, 255)

            self.update_Synapse_Historgrams(Network_UI, group, net_color_input)



from PymoNNto.Exploration.Network_UI.TabBase import *

class scatter_tab(TabBase):

    def __init__(self, x_var, y_var, title='Scatter', timesteps=500):
        super().__init__(title)
        self.x_var = x_var
        self.y_var = y_var
        self.timesteps=timesteps

    def add_recorder_variables(self, neuron_group, Network_UI):
        Network_UI.add_recording_variable(neuron_group, self.x_var, timesteps=self.timesteps)
        Network_UI.add_recording_variable(neuron_group, self.y_var, timesteps=self.timesteps)

    def initialize(self, Network_UI):
        self.scatter_tab = Network_UI.add_tab(self.title)

        p = Network_UI.tab.add_plot(x_label=self.x_var, y_label=self.y_var)
        self.scatter_items=[]
        for i, _ in enumerate(Network_UI.get_visible_neuron_groups()):
            spi = pg.ScatterPlotItem()
            self.scatter_items.append(spi)
            p.addItem(spi)

    def update(self, Network_UI):
        if self.scatter_tab.isVisible():
            for i, group in enumerate(Network_UI.get_visible_neuron_groups()):

                try:
                    x_values = group[self.x_var, 0, 'np'][-self.timesteps:]
                    y_values = group[self.y_var, 0, 'np'][-self.timesteps:]

                    x_val = np.mean(x_values, axis=1)
                    y_val = np.mean(y_values, axis=1)

                    self.scatter_items[i].setData(x_val.copy(), y_val.copy(), brush=group.color)
                except:
                    pass



from PymoNNto.Exploration.Evolution.common_UI import *
from PymoNNto.Exploration.Evolution.communication import *
from tkinter import Tk  #for clipboard access

class Evolution_UI(Execution_Manager_UI_Base):

    def get_folder(self):
        return 'Evolution_Project_Clones'

    def get_title(self):
        return 'Evolution Manager'



    def add_ui_elements(self, left_vertical_layout, right_vertical_layout):
        #horizontal_layout = QHBoxLayout()
        #self.sidebar_current_vertical_layout.addLayout(horizontal_layout)

        #left_vertical_layout = QVBoxLayout()
        #horizontal_layout.addLayout(left_vertical_layout)
        label = QLabel(u"\u24D8 " + "Slave file") #\u24BE
        label.setToolTip("path to python file which is supposed to be evolved (relative from root folder in which the Data folder is located)")
        left_vertical_layout.addWidget(label)

        label = QLabel(u"\u24D8 " + 'Thread number')
        label.setToolTip("number of worker threads")
        left_vertical_layout.addWidget(label)

        label = QLabel(u"\u24D8 " + 'Individual count')
        label.setToolTip("number of individuals in one generation")
        left_vertical_layout.addWidget(label)

        label = QLabel(u"\u24D8 " + 'Mutation')
        label.setToolTip("the parameter determines the mutation with: gene_new = normal(gene_old, clip(gene_old*mutation,0))")
        left_vertical_layout.addWidget(label)

        label = QLabel(u"\u24D8 " + 'Death rate')
        label.setToolTip("number of individuals with bad scores which are removed from each generation")
        left_vertical_layout.addWidget(label)

        #pc_label = QLabel(u"\u24D8 " + 'Python cmd')
        #pc_label.setToolTip("python command of the execution device (python or python3)")
        #left_vertical_layout.addWidget(pc_label)

        ii_label = QLabel(u"\u24D8 " + 'Inactive info')
        ii_label.setToolTip("additional genome that is passed to individuals but is not changed or optimized during the evolution run")
        left_vertical_layout.addWidget(ii_label)

        c_label = QLabel(u"\u24D8 " + 'Constriants')
        c_label.setToolTip("Example: ['a<b','b<=c'] if a b and c are genes")
        left_vertical_layout.addWidget(c_label)

        eo_label = QLabel(u"\u24D8 " + 'Evo options')
        eo_label.setToolTip("not used for now")
        left_vertical_layout.addWidget(eo_label)


        self.slave_file_edit = QLineEdit('')
        self.slave_file_edit.textChanged.connect(self.check_file)
        self.slave_file_edit.setText('Exploration/Evolution/example_slave.py')

        self.thread_number_edit = QLineEdit('4')
        #self.python_cmd_edit = QLineEdit("python3")
        #self.python_cmd_edit = QComboBox()
        #self.python_cmd_edit.addItems(["python3", "python"])
        self.individual_count_edit = QLineEdit('10')
        self.mutation_edit = QLineEdit('0.05')
        self.death_rate_edit = QLineEdit('0.5')
        #self.start_genomes_edit.setFixedHeight(200)
        # QLineEdit("[{'a':1,'b':2,'c':2,'d':2,'e':3}]")
        self.inactive_genome_info_edit = QLineEdit('{}')
        self.constraints_edit = QLineEdit("[]")#'a<b','b<=c'
        self.evo_options_edit = QLineEdit('{}')

        #right_vertical_layout = QVBoxLayout()
        #horizontal_layout.addLayout(right_vertical_layout)

        file_horizontal_layout = QHBoxLayout()
        file_horizontal_layout.addWidget(self.slave_file_edit, stretch=10)

        file_btn = QPushButton("...")
        file_btn.clicked.connect(self.select_file)
        file_btn.setMaximumWidth(30)
        file_horizontal_layout.addWidget(file_btn, stretch=0)


        right_vertical_layout.addLayout(file_horizontal_layout)

        #right_vertical_layout.addWidget(self.slave_file_edit)
        right_vertical_layout.addWidget(self.thread_number_edit)
        right_vertical_layout.addWidget(self.individual_count_edit)
        right_vertical_layout.addWidget(self.mutation_edit)
        right_vertical_layout.addWidget(self.death_rate_edit)
        #right_vertical_layout.addWidget(self.python_cmd_edit)
        right_vertical_layout.addWidget(self.inactive_genome_info_edit)
        right_vertical_layout.addWidget(self.constraints_edit)
        right_vertical_layout.addWidget(self.evo_options_edit)

        more_info_label = self.sidebar.add_widget(QLabel("..."), stretch=1)

        label = self.sidebar.add_widget(QLabel(u"\u24D8 " + "Start genomes:"), stretch=1)
        label.setToolTip(
            "Initial individuals and their genes.")


        self.sidebar.add_row()

        extract_btn = self.sidebar.add_widget(QPushButton('extract default genome'))
        extract_btn.clicked.connect(self.extract_genome)

        add_ind_btn = self.sidebar.add_widget(QPushButton('add individual '+u"\u24D8"))
        add_ind_btn.setToolTip('Adds an empty row or genomes that are copied to the clipboard.')
        add_ind_btn.clicked.connect(self.add_individual)

        add_gene_btn = self.sidebar.add_widget(QPushButton('add gene'))
        add_gene_btn.clicked.connect(self.add_gene)

        self.sidebar.set_parent_layout()

        self.gene_table = self.sidebar.add_widget(QGeneTableWidget())


        #txt = QTextEdit()
        #txt.setText("""[
        #{'a':1,'b':2,'c':2}
        #]""")
        #self.start_genomes_edit = self.sidebar.add_widget(txt, stretch=100)


        #right_vertical_layout.addWidget(self.start_genomes_edit, stretch=5)

        more_info_label.setVisible(True)
        #pc_label.setVisible(False)
        ii_label.setVisible(False)
        c_label.setVisible(False)
        eo_label.setVisible(False)
        #self.python_cmd_edit.setVisible(False)
        self.inactive_genome_info_edit.setVisible(False)
        self.constraints_edit.setVisible(False)
        self.evo_options_edit.setVisible(False)

        def more_clicked(event):
            more_info_label.setVisible(False)
            #pc_label.setVisible(True)
            ii_label.setVisible(True)
            c_label.setVisible(True)
            #eo_label.setVisible(True)
            #self.python_cmd_edit.setVisible(True)
            self.inactive_genome_info_edit.setVisible(True)
            self.constraints_edit.setVisible(True)
            #self.evo_options_edit.setVisible(True)


        more_info_label.mouseReleaseEvent = more_clicked
        #more_info_label.clicked.connect(more_clicked)

    def add_individual(self):
        #self.text_input_dialog(txt, 'genomes (leave empty for empty ): ', func, default_text='[{}]')
        genomes=[{}]
        try:
            clipboard = Tk().clipboard_get()
            if clipboard[0]=='[' and clipboard[-1]==']':
                genomes=eval(clipboard)
            if clipboard[0]=='{' and clipboard[-1]=='}':
                genomes=[eval(clipboard)]
        except:
            pass
        self.gene_table.add_genomes(genomes)

    def add_gene(self):
        gene_name = self.text_input_dialog('Gene name:', 'add gene', self.__add_gene)

    def __add_gene(self, gene_name):
        if len(gene_name)>0:
            self.gene_table.add_gene(gene_name)

    def get_genomes_from_file(self, file):
        print_output = execute_local_file(file, 'None', 0, get_gene_mode=True, get_output=True)

        print(print_output)

        result = []
        result.append({})

        for line in print_output.split('\n'):
            # if '#get_gene#' in line:
            ls = line.split('#')

            if len(ls) == 3 and (ls[1] == 'set_genome' or ls[1] == 'genome_set'):
                if len(result[-1])>0:
                    result.append({})

            if len(ls) == 6 and ls[1] == 'get_gene':
                gene = ls[2]
                default = ls[3]
                result[-1][gene] = default

                #type = ls[4]
                #result.append([gene, default, type])
                #print(gene, default, type)

        return result

    def extract_genome(self):
        file = get_root_folder()+'/'+self.slave_file_edit.text()
        default_genomes = self.get_genomes_from_file(file)

        #default_genes = self.get_genome_from_file(get_root_folder()+'/'+self.slave_file_edit.text())
        #default_genome = {g[0]:g[1] for g in default_genes}
        #test_gemone = {'a':1, 'b':2, 'c':100, 'd':9}
        #test_gemone2 = {'a': 2, 'b': 2, 'x': 999, 'c': 100, 'd': 9}
        #test_gemone3 = {'a': 1}

        self.gene_table.add_genomes(default_genomes)

    def On_Tab_Changed(self, i):
        if i is not None:
            tab = self.tabs.currentWidget()
            if hasattr(tab, 'ssm'):
                #self.update_status(tab, tab.name)
                ssm = self.tabs.currentWidget().ssm
                self.set_text(ssm, self.task_name_edit, 'name')
                self.set_text(ssm, self.slave_file_edit, 'slave_file')
                self.set_text(ssm, self.inactive_genome_info_edit, 'inactive_genome_info')
                self.set_text(ssm, self.thread_number_edit, 'thread_number')
                self.set_text(ssm, self.individual_count_edit, 'individual_count')
                self.set_text(ssm, self.mutation_edit, 'mutation')
                self.set_text(ssm, self.death_rate_edit, 'death_rate')

                #self.set_text(ssm, self.start_genomes_edit, 'start_genomes')
                self.gene_table.clear()
                data = ssm.load_param('start_genomes', default='', return_string=True)
                if data is not None and data != '':
                    print(data)
                    self.gene_table.add_genomes(eval(data))

                self.set_text(ssm, self.constraints_edit, 'constraints')
                #self.set_text(ssm, self.python_cmd_edit, 'python_cmd')
                self.set_text(ssm, self.evo_options_edit, 'evo_options')
                #self.refresh_view(tab)

    def get_help_txt(self):
        return '''Every tab corresponds to a folder in: Data/Evolution_Project_Clones/
This folders also contains the file execute.py which is used to start the evolution.
All genomes generate a file with the results and the genes when set_score is called in the slave file.
      
      
        
Local:
    The project folder (in which the "Data" folder is located) is copied into Data/Evolution_Project_Clones/task_name/



Remote Servers:
    For remote SSH servers the local folder (Data/Evolution_Project_Clones/task_name/) only contains some config files, 
    while the copy of the project folder is located in the user folder of the server.
    When the refresh button is clicked, the genome files are transferred back to the local folder.
    
    
    Requirements for remote server (tested on Ubuntu on AWS):
        -Python            (3)
        -screen
            
        -pip               (sudo apt install python3-pip)
        -PymoNNto          (pip install PymoNNto)
            
        -zip               (for transfer to remote server | sudo apt install zip)
        -unzip             (for transfer back to controller | sudo apt install unzip)
        -libgl1-mesa-glx   (if you have PyQT5 elements imported | sudo apt install libgl1-mesa-glx)
            
    Make sure that the python3 (python3 file.py) command is working on your system. "python file.py" (usually on windows) 
    is not supported. (create copy of python.exe named python3.exe if this problem occurs)
            
    When the evolution is started on a ssh server it creates a "screen" session which is named after the evolution name.

    Useful functions are:
        "screen -list" to show the screen sessions and their ids,
        "screen -r id" to connect to a session where "id" has to be replaced and
        "exit" to close a screen session you are connected to.
'''

    def create_execution_file(self, name, folder, file):
        exec_file = """
from PymoNNto.Exploration.Evolution.Evolution import *
if __name__ == '__main__':
    evo = Evolution(name='#name#',
                slave_file='#slave_file#',
                individual_count=#individual_count#,
                mutation=#mutation#,
                death_rate=#death_rate#,
                constraints=#constraints#,
                inactive_genome_info=#inactive_genome_info#,
                start_genomes=#start_genomes#,
                devices={'multi_thread': #thread_number#},
                additional_evo_params=#evo_options#
                )

    if not evo.start(ui=False):
        evo.continue_evolution(ui=False)
        """

        print('generate execute.py...')
        exec_file = exec_file.replace('#name#', name)
        exec_file = exec_file.replace('#slave_file#', self.slave_file_edit.text())
        exec_file = exec_file.replace('#inactive_genome_info#', self.inactive_genome_info_edit.text())
        exec_file = exec_file.replace('#thread_number#', self.thread_number_edit.text())
        exec_file = exec_file.replace('#individual_count#', self.individual_count_edit.text())
        exec_file = exec_file.replace('#mutation#', self.mutation_edit.text())
        exec_file = exec_file.replace('#death_rate#', self.death_rate_edit.text())
        exec_file = exec_file.replace('#start_genomes#', str(self.gene_table.get_genomes()))#self.start_genomes_edit.toPlainText().replace('\r', '').replace('\n', '')
        exec_file = exec_file.replace('#constraints#', self.constraints_edit.text())
        exec_file = exec_file.replace('#evo_options#', self.evo_options_edit.text())

        md_file = open(file, "w")
        md_file.write(exec_file)
        md_file.close()
        return True

    def save_configuration(self, ssm):#SimpleStorageManager

        ssm.save_param('slave_file', self.slave_file_edit.text())
        ssm.save_param('inactive_genome_info', self.inactive_genome_info_edit.text())
        ssm.save_param('thread_number', self.thread_number_edit.text())
        ssm.save_param('individual_count', self.individual_count_edit.text())
        ssm.save_param('mutation', self.mutation_edit.text())
        ssm.save_param('death_rate', self.death_rate_edit.text())
        ssm.save_param('start_genomes', str(self.gene_table.get_genomes()))#self.start_genomes_edit.toPlainText().replace('\r', '').replace('\n', '')
        ssm.save_param('constraints', self.constraints_edit.text())
        #ssm.save_param('python_cmd', self.python_cmd_edit.text())
        ssm.save_param('evo_options', self.evo_options_edit.text())

        gene_keys = self.gene_table.get_genes() #list(eval(self.start_genomes_edit.toPlainText().replace('\r', '').replace('\n', ''))[0].keys())
        ssm.save_param('gene_keys', gene_keys)

    def valid_configuration(self):
        #valid_genomes = False
        #try:
        #    sg = eval(self.start_genomes_edit.toPlainText().replace('\r', '').replace('\n', ''))
        #    for key in ['generation', 'score', 'id']:
        #        for genome in sg:
        #            if key in genome:
        #                genome.pop(key)
        #                print(key, 'removed from genomes')
        #    self.start_genomes_edit.setText(str(sg))
        #    valid_genomes = True
        #except:
        #    print('error parsing start genomes')

        #if not valid_genomes:
       #     print('error parsing start genomes')

        valid = self.gene_table.check_cells()

        if not valid:
            self.show_message('Error', 'cannot parse genomes', 'ok')

        return valid

    def add_additional_tab_elements(self, tab, name):

        '''
        def send(genome):
            return
            #TODO
            #user, host, password = split_ssh_user_host_password_string(tab.server,False)
            #if host is None:
            #    host = 'localhost'
            #print(socket_send(host, 'insert\r\n' + genome))

        def insert_click(event):
            self.text_input_dialog('enter genome', 'send', send, default_text='{}')


        insert_btn = self.tab.add_widget(QPushButton("insert genome"))
        insert_btn.clicked.connect(insert_click)
        '''

        self.tab.add_row(stretch=10)
        #self.Next_H_Block(stretch=10)
        add_evolution_plot_items(self, tab)
        tab.folder = get_epc_folder(self.folder) + '/' + name + '/'

    def refresh_view(self, tab):
        if tab.gene_keys is not None:
            update_evolution_plot(self, tab, tab.name, tab.gene_keys,data_folder=get_data_folder() + '/' + self.folder + '/' + tab.name + '/Data')


class QGeneTableWidget(QTableWidget):

    def __init__(self):
        super().__init__()
        self.clear()
        self.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        #self.verticalHeader().hide()
        #self.cellClicked.connect(self.cell_clicked)
        self.cellChanged.connect(self.cell_changed)

        self.horizontalHeader().sectionClicked.connect(self.horizontal_header_clicked)
        self.verticalHeader().sectionClicked.connect(self.vertical_header_clicked)

    def clear(self):
        self.clearContents()
        self.setRowCount(0)
        self.setColumnCount(0)

    def get_genomes(self):
        check = self.check_cells()
        if check:
            result = []
            for r in range(self.rowCount()):
                genome = {}
                for c in range(self.columnCount()):
                    gene_key = self.horizontalHeaderItem(c).text()
                    gene_value = float(self.item(r, c).text())
                    genome[gene_key] = gene_value
                result.append(genome)
            return result
        else:
            return None


    def check_cells(self):
        check = True
        for r in range(self.rowCount()):
            for c in range(self.columnCount()):
                item = self.item(r, c)
                if item is None:
                    item=QTableWidgetItem()
                    self.setItem(r, c, item)
                if not is_number(item.text()):
                    item.setBackground(QColor(255,150,150))
                    check=False
                else:
                    item.setBackground(QColor("white"))
        return check


    def cell_changed(self, row, column):
        self.check_cells()

    #def cell_clicked(self, row, column):
    #    print(self.get_genomes())
    #    print("Row %d and Column %d was clicked " % (row, column))
    #    item = self.item(row, column)
    #    print(item.text())

    def horizontal_header_clicked(self, i):
        self.removeColumn(i)

    def vertical_header_clicked(self, i):
        self.removeRow(i)


    def add_genomes(self, genomes):
        for genome in genomes:
            self.add_genome(genome)

    def add_genome(self, genome):
        rc=self.rowCount()
        self.insertRow(rc)
        self.setVerticalHeaderItem(rc, QTableWidgetItem('x'))
        for gene_key, gene_value in genome.items():
            self.row_insert(gene_key, gene_value)
        self.check_cells()

    def row_insert(self, key, value):
        col_index = self.get_column(key, True)
        row_index = self.rowCount()-1
        self.setItem(row_index, col_index, QTableWidgetItem(str(value)))

    def get_column(self, key, append_if_missing=False):
        for i, g in enumerate(self.get_genes()):
            if key==g:
                return i

        if append_if_missing:
            return self.add_gene(key)

        return -1

    def add_gene(self, key):
        cc=self.columnCount()
        self.insertColumn(cc)
        self.setHorizontalHeaderItem(cc, QTableWidgetItem(str(key)))
        self.check_cells()
        return cc

    def get_genes(self):
        #print(self.getHorizontalHeaderLabels)
        return [self.horizontalHeaderItem(c).text() for c in range(self.columnCount())]#self.verticalHeaderItems()

    #def dsfads(self):
    #    self.gene_table.setRowCount(0)
    #    self.gene_table.setColumnCount(len(default_genes)+1)
    #    self.gene_table.setHorizontalHeaderLabels([g[0] for g in default_genes]+['+'])

    #    for r, genome in enumerate([default_genome, [2,2,1,3], [7,2,3,9]]):
    #        #QTableWidget.insertRow(0)
    #        self.gene_table.insertRow(self.gene_table.rowCount())
    #        for c,g in enumerate(genome+['remove']):
    #            self.gene_table.setItem(r, c, QTableWidgetItem(str(g)))

    #    self.gene_table.insertRow(self.gene_table.rowCount())




########################################################### Exception handling

def except_hook(cls, exception, traceback):
    sys.__excepthook__(cls, exception, traceback)

sys.excepthook = except_hook

if __name__ == '__main__':
    Evolution_UI().show()




from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.NetworkCore.Base_Attachable_Modules import *
from PymoNNto.Exploration.StorageManager.StorageManager import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.info_tabs import *
from PymoNNto.Exploration.HelperFunctions.Save_Load import *
import copy
import pickle
import os

class sidebar_save_load_module(TabBase):

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):

        self.folder=get_data_folder()+'/NetworkStates/'

        if not os.path.exists(self.folder):
            try:
                os.mkdir(self.folder)
            except:
                print('was not able to create Data/NetworkStates folder')

        Network_UI.sidebar.add_row()

        self.save_edit = Network_UI.sidebar.add_widget(QLineEdit('name...'))

        def save(event):
            save_network(Network_UI.network, self.save_edit.text())
            #net = copy.deepcopy(Network_UI.network)
            #clear_compiled_code(net)
            #pickle.dump(net, open(self.folder + self.save_edit.text() + '.netstate', 'wb'))
            self.load_box.addItem(self.save_edit.text())

        self.save_btn = Network_UI.sidebar.add_widget(QPushButton('save', Network_UI.main_window))
        self.save_btn.clicked.connect(save)


        self.load_box = Network_UI.sidebar.add_widget(QComboBox())
        for file in os.listdir(self.folder):
            if '.netstate' in str(file):
                self.load_box.addItem(file.replace('.netstate', ''))


        def load(event):
            Network_UI.network = load_network(self.load_box.currentText())

            #Network_UI.network = pickle.load(open(self.folder + self.load_box.currentText() + '.netstate', 'rb'))

            for tab in Network_UI.infotabs:
                Network_UI.tabs.removeTab(len(Network_UI.tabs) - 1)

            Network_UI.modules.append(info_tab(Network_UI))

        self.load_btn = Network_UI.sidebar.add_widget(QPushButton('load', Network_UI.main_window))
        self.load_btn.clicked.connect(load)


    def update(self, Network_UI):
        return


from PymoNNto.NetworkCore.Base_Tagable_Object import *
import numpy as np
from numpy.random import *
from numpy import *
from functools import wraps

def lognormal_real_mean(mean=1.0, sigma=1.0, size=1):
    mu = -np.power(sigma, 2) + np.log(mean)
    return lognormal(mu, sigma, size=size)

def lognormal_rm(mean=1.0, sigma=1.0, size=1):
    return lognormal_real_mean(mean,sigma,size)

def uniform_gap(mean=1.0, gap_percent=10, size=1):
    return uniform(low=mean-mean*gap_percent, high=mean+-mean*gap_percent, size=size)

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def deprecated_warning(message):
    def inner_function(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            #raise Exception(message)
            print(message)
            return function(*args, **kwargs)
        return wrapper
    return inner_function

class NetworkObjectBase(TaggableObjectBase):

    def __init__(self, tag, network, behavior):
        super().__init__(tag)

        self.network = network
        self.behavior = {}#behavior

        if type(behavior) == list:
            self.behavior = dict(zip(range(len(self.behavior)), self.behavior))

        for k,b in behavior.items():
            self.add_behavior(k, b, initialize=False)

        # quick access
        #for b in self.behavior.values():
        #    for tag in b.tags:
        #        if not hasattr(self, tag):
        #            setattr(self, tag, b)

        #for k,b in self.behavior.items():
        #    self.network._add_behavior_to_sorted_execution_list(k, self, b)

        for k in sorted(list(self.behavior.keys())):
            if self.behavior[k].initialize_on_init:
                self.behavior[k].initialize(self)
                #behavior.check_unused_attrs()
                #network._initialize_check(self, k)

        #self.learning = True
        self.recording = True

        self.analysis_modules = []

    def add_behaviors(self, behavior_dict):
        for key in behavior_dict:
            self.add_behavior(key, behavior_dict[key])
        return behavior_dict



    def add_behavior(self, key, behavior, initialize=True):
        #check key already exists!!!
        if not key in self.behavior:
            self.behavior[key] = behavior
            self.network._add_behavior_to_sorted_execution_list(key, self, self.behavior[key])
            self.network._add_key_to_sorted_behavior_timesteps(key)#remove!!!
            self.network.clear_tag_cache()
            if initialize:
                #behavior.initialize_init(self)
                behavior.initialize(self)
                #behavior.initialize_last(self)
                behavior.check_unused_attrs()

            #quick access
            for tag in behavior.tags:
                if not hasattr(self, tag):
                    setattr(self, tag, behavior)

            return behavior
        else:
            raise Exception('Error: Key already exists.'+str(key))

    def remove_behavior(self, key_tag_behavior_or_type):
        remove_keys=[]
        for key in self.behavior:
            b = self.behavior[key]
            if key_tag_behavior_or_type == key or key_tag_behavior_or_type in b.tags or key_tag_behavior_or_type == b or key_tag_behavior_or_type == type(b):
                remove_keys.append(key)
        for key in remove_keys:
            b=self.behavior.pop(key)
            self.network._remove_behavior_from_sorted_execution_list(key, self, b)
            #remove quick access
            for tag in b.tags:
                if hasattr(self, tag) and getattr(self, tag)==b:
                    delattr(self, tag)

    def set_behaviors(self, tag, enabeled):
        if enabeled:
            print('activating', tag)
        else:
            print('deactivating', tag)
        for b in self[tag]:
            b.behavior_enabled = enabeled


    def deactivate_behaviors(self, tag):
        self.set_behaviors(tag, False)

    def activate_behaviors(self, tag):
        self.set_behaviors(tag, True)


    def find_objects(self, key):
        result = []

        if key in self.behavior:
            result.append(self.behavior[key])

        for bk in self.behavior:
            behavior = self.behavior[bk]
            result += behavior[key]

        for am in self.analysis_modules:
            result += am[key]

        return result

    def add_analysis_module(self, module):
        module._attach_and_initialize_(self)

    def get_all_analysis_module_results(self, tag, return_modules=False): #...get_analysis_module_results('Classification')
        result = {}
        modules = {}
        for module in self[tag]:
            module_results = module.get_results()
            for k in module_results:
                result[k] = module_results[k]
                modules[k] = module
        if return_modules:
            return result, modules
        else:
            return result


    def _get_mat(self, mode, dim, density=None, scale=None, plot=False):
        if mode not in self._mat_eval_dict:
            '''
            ev_str = mode

            cast = True

            if ev_str == bool or ev_str == 'bool':
                cast = False
                ev_str = 'zeros(dtype=bool)'

            if ev_str == int or ev_str == 'int':
                cast = False
                ev_str = 'zeros(dtype=int)'

            if ev_str == 'random' or ev_str == 'rand' or ev_str == 'rnd':
                ev_str = 'uniform'

            if type(ev_str) == int or type(ev_str) == float:
                ev_str = str(ev_str)+'*ones()'

            if '(' not in ev_str and ')' not in ev_str:
                ev_str += '()'

            if 'zeros' in ev_str or 'ones' in ev_str:
                a1 = 'shape=dim'
            else:
                a1 = 'size=dim'

            ev_str = ev_str.replace(')', ',' + a1 + ')')
            ev_str = ev_str.replace('(,', '(')

            if cast:
                ev_str += '.astype(self.def_dtype)'

            self._mat_eval_dict[mode] = compile(ev_str, '<string>', 'eval')
            '''
            self._mat_eval_dict[mode] = compile(self.network.backend.evaluate_string(mode), '<string>', 'eval')

        result = eval(self._mat_eval_dict[mode])
        return self.network.backend.modulate(result, dim, density, scale, plot)

    def get_nparray(self, dim):
        return np.zeros(dim).astype(self.def_dtype)

    def get_buffer_mat(self, dim, size):
        return np.array([self.get_nparray(dim) for _ in range(size)])

    def buffer_roll(self, mat, new=None, counter=False):
        #return np.roll(mat, 1, axis=0)
        mat[1-counter : len(mat)-counter] = mat[0+counter : len(mat)-1+counter]

        if new is not None:
            mat[0-counter]=new

        return mat

    ################################################################################################
    #deprecated#####################################################################################
    ################################################################################################

    @deprecated_warning("set_mechanisms function will be removed in coming versions. Please use set_behaviors instead.")
    def set_mechanisms(self, tag):
        if tag is str:
            tag=[tag]
        for t in tag:
            set_behaviors(t)

    @deprecated_warning("deactivate_mechanisms function will be removed in coming versions. Please use deactivate_behaviors instead.")
    def deactivate_mechanisms(self, tag):
        self.set_behaviors(tag, False)

    @deprecated_warning("activate_mechanisms function will be removed in coming versions. Please use activate_behaviors instead.")
    def activate_mechanisms(self, tag):
        self.set_behaviors(tag, True)

    @deprecated_warning("add_behaviours function will be removed in coming versions. Please use add_behaviors instead.")
    def add_behaviours(self, behavior_dict):
        return self.add_behaviors(behavior_dict)

    @deprecated_warning("add_behaviour function will be removed in coming versions. Please use add_behavior instead.")
    def add_behaviour(self, key, behavior, initialize=True):
        return self.add_behavior(key, behavior, initialize)

    @deprecated_warning("remove_behaviour function will be removed in coming versions. Please use remove_behavior instead.")
    def remove_behaviour(self, key_tag_behavior_or_type):
        return self.remove_behavior(key_tag_behavior_or_type)

    @deprecated_warning("set_behaviours function will be removed in coming versions. Please use set_behaviors instead.")
    def set_behaviours(self, tag, enabeled):
        return self.set_behaviors(tag, enabeled)

    @deprecated_warning("deactivate_behaviours function will be removed in coming versions. Please use deactivate_behaviors instead.")
    def deactivate_behaviours(self, tag):
        return self.deactivate_behaviors(tag)

    @deprecated_warning("activate_behaviours function will be removed in coming versions. Please use activate_behaviors instead.")
    def activate_behaviours(self,tag):
        return self.activate_behaviors(tag)


from time import time
import numpy as np
from PymoNNto.NetworkCore.Base_Attachable_Modules import *
from PymoNNto.NetworkCore.Synapse_Group import *
from PymoNNto.Exploration.Evolution.Interface_Functions import *
import copy
import time
import sys
from PymoNNto.NetworkCore.NumPy_Backend.NumPy_Backend import *

float32 = np.float32
float64 = np.float64

SxD = 0
DxS = 1

class Network(NetworkObjectBase):

    def __init__(self, tag=None, behavior={}, settings={}):
        self.apply_settings(settings)

        self.NeuronGroups = []
        self.SynapseGroups = []

        self.iteration = 0

        self.behavior_timesteps = []
        self.sorted_behavior_execution_list = [] #stores (key, beh_parent, behavior) triplets
        
        super().__init__(tag, self, behavior)

    # {'dtype':float32, 'syn_dim':DxS}
    def apply_settings(self, settings):
        #backend_class = settings.get('backend', NumPy)
        #self.backend = backend_class()
        self.backend = NumPy()
        self.def_dtype = settings.get('dtype', float32)
        self.transposed_synapse_matrix_mode = settings.get('syn_dim', DxS)!=DxS

    def all_objects(self):
        return [self]+self.NeuronGroups+self.SynapseGroups

    def _add_key_to_sorted_behavior_timesteps(self, key):
        if key not in self.behavior_timesteps:
            self.behavior_timesteps.append(key)
            self.behavior_timesteps.sort()


    def _add_behavior_to_sorted_execution_list(self, key, beh_parent, behavior):
        insert_indx=0
        for i,kpb in enumerate(self.sorted_behavior_execution_list):
            k, p, b = kpb
            if key>=k:
                insert_indx = i+1
        self.sorted_behavior_execution_list.insert(insert_indx, (key, beh_parent, behavior))
        #print([k for k,_,_ in self.sorted_behavior_execution_list])
        
    def _remove_behavior_from_sorted_execution_list(self, key, beh_parent, behavior):
        rm_indx = -1
        for i,kpb in enumerate(self.sorted_behavior_execution_list):
            k, p, b = kpb
            if key==k and beh_parent==p and behavior==b:
                rm_indx = i
                break
        if rm_indx>-1:
            self.sorted_behavior_execution_list.pop(rm_indx)
        else:
            raise Exception('behavior not found')

    #######################################
    #Behavior Management
    #######################################

    def set_behaviors(self, tag, enabeled):
        if enabeled:
            print('activating', tag)
        else:
            print('deactivating', tag)
        for obj in self.all_objects():
            for b in obj[tag]:
                b.behavior_enabled = enabeled

    def all_behaviors(self):
        result = []
        for obj in self.all_objects():
            for beh in obj.behavior.values():
                result.append(beh)
        return result

    #######################################
    #Recorder controll
    #######################################

    def recording_off(self):
        for obj in self.all_objects():
            obj.recording = False

    def recording_on(self):
        for obj in self.all_objects():
            obj.recording = True

    def clear_recorder(self, keys=None):
        for obj in self.all_objects():
            for key in obj.behavior:
                if (keys is None or key in keys) and hasattr(obj.behavior[key], 'clear_recorder'):
                    obj.behavior[key].clear_recorder()


    #######################################
    #Initialization
    #######################################

    def initialize(self, info=True, warnings=True, storage_manager=None):
        self.set_gene_variables(info=info, storage_manager=storage_manager)

        if info:
            desc=str(self)
            print(desc)
            if storage_manager is not None:
                storage_manager.save_param('info', desc)

        #self.set_synapses_to_neuron_groups()

        self.initialize_behaviors()
        self.check_unique_tags(warnings)

    #def _initialize_check(self, obj, key):##set as decorator
    #    obj_keys_before = list(obj.__dict__.keys())
    #    beh_keys_before = list(obj.behavior[key].__dict__.keys())
    #    obj.behavior[key].initialize(obj)
    #    obj_keys_after = list(obj.__dict__.keys())
    #    beh_keys_after = list(obj.behavior[key].__dict__.keys())
    #    obj.behavior[key]._created_obj_variables = list(set(obj_keys_after) - set(obj_keys_before))
    #    obj.behavior[key]._created_beh_variables = list(set(beh_keys_after) - set(beh_keys_before))

    def initialize_behaviors(self):

        for key, parent, behavior in self.sorted_behavior_execution_list:
            if not behavior.initialize_on_init and not behavior.initialize_last:
                behavior.initialize(parent)
                behavior.check_unused_attrs()

        for key, parent, behavior in self.sorted_behavior_execution_list:
            if behavior.initialize_last:
                behavior.initialize(parent)
                behavior.check_unused_attrs()
                
        for key, parent, behavior in self.sorted_behavior_execution_list:# add quick access
            for tag in behavior.tags:
                if not hasattr(parent, tag):
                    setattr(parent, tag, behavior)


    #######################################
    #Initialization helper
    #######################################

    def set_gene_variables(self, info=True, storage_manager=None):

        if get_gene_mode():
            sys.exit()

        for obj in self.all_objects():
            for key in obj.behavior:
                b = obj.behavior[key]
                b.set_gene_variables()

        if info:
            genome = get_genome()
            if len(genome)>0:
                print('genome:', genome)

        if storage_manager is not None:
            storage_manager.save_param(key='evolution_params', value=get_genome())
            storage_manager.save_param_dict(get_genome())


    def __str__(self):
        neuron_count = np.sum(np.array([ng.size for ng in self.NeuronGroups]))
        synapse_count = np.sum(np.array([sg.src.size*sg.dst.size for sg in self.SynapseGroups]))

        basic_info = '(Neurons: '+str(neuron_count)+'|'+str(len(self.NeuronGroups))+' groups, Synapses: '+str(synapse_count)+'|'+str(len(self.SynapseGroups))+' groups)'

        result = 'Network'+str(self.tags)+basic_info+'{'
        for k in sorted(list(self.behavior.keys())):
            result += str(k)+':'+str(self.behavior[k])
        result += '}'+'\r\n'

        for ng in self.NeuronGroups:
            result += str(ng)+'\r\n'

        used_tags = []
        for sg in self.SynapseGroups:
            tags = str(sg.tags)
            if tags not in used_tags:
                result += str(sg)+'\r\n'
            used_tags.append(tags)

        return result[:-2]

    def check_unique_tags(self,warnings=True):
        unique_tags=[]
        for ng in self.NeuronGroups:

            if len(ng.tags) == 0:
                ng.tags.append('NG')
                print('no tag defined for NeuronGroup. "NG" tag added')

            if ng.tags[0] in unique_tags:
                counts=unique_tags.count(ng.tags[0])
                new_tag=ng.tags[0]+chr(97+counts)
                unique_tags.append(ng.tags[0])
                if warnings:
                    print('Warning: NeuronGroup Tag "'+ng.tags[0]+'" already in use. The first Tag of an Object should be unique and will be renamed to "'+new_tag+'". Multiple Tags can be separated with a "," (NeuronGroup(..., tag="tag1,tag2,..."))')
                ng.tags[0] = new_tag

            else:
                unique_tags.append(ng.tags[0])


    #######################################
    #Iteration simulation
    #######################################

    def simulate_iteration(self, measure_behavior_execution_time=False):

        if measure_behavior_execution_time:
            if not hasattr(self, 'time_measures'):
                self.time_measures={key:0.0 for key, _, _ in self.sorted_behavior_execution_list}

        self.iteration += 1

        for key, parent, behavior in self.sorted_behavior_execution_list:
            if behavior.behavior_enabled and not behavior.empty_iteration_function:
                if measure_behavior_execution_time:
                    start_time = time.time()
                    behavior.iteration(parent)
                    self.time_measures[key] += (time.time() - start_time) * 1000
                else:
                    behavior.iteration(parent)

        if measure_behavior_execution_time:
            return self.time_measures


    def simulate_iterations(self, iterations, batch_size=100, measure_block_time=True, disable_recording=False, batch_progress_update_func=None):

        if type(iterations) is str:
            iterations=self['Clock', 0].time_to_iterations(iterations)

        if type(batch_size) is str:
            batch_size=self['Clock', 0].time_to_iterations(batch_size)

        time_diff=None

        if disable_recording:
            self.recording_off()


        if batch_size==-1:
            outside_it = 1
            block_iterations = iterations
        else:
            outside_it=int(iterations/batch_size)
            block_iterations=batch_size

        out_it=range(int(outside_it))

        for t in range(int(outside_it)):
            if measure_block_time:
                start_time = time.time()
            for i in range(int(block_iterations)):
                self.simulate_iteration()
            if measure_block_time:
                time_diff = (time.time() - start_time) * 1000
                remaining = time_diff * (outside_it-t) / 1000 / 60
                print('\r{}x {}/{} ({}%) {}ms {:.1f}min'.format(block_iterations,t+1, outside_it, int(100/outside_it*(t+1)), int(time_diff), remaining), end='')#, end='')

            if batch_progress_update_func is not None:
                batch_progress_update_func((t+1.0)/int(outside_it)*100.0, self)

        for i in range(iterations%batch_size):
            self.simulate_iteration()

        if disable_recording:
            self.recording_on()

        if measure_block_time:
            print('')

        return time_diff

    #######################################
    # Tagging
    #######################################

    def find_objects(self, key):
        result = super().find_objects(key)

        for ng in self.NeuronGroups:
            result += ng[key]

        for sg in self.SynapseGroups:
            result += sg[key]

        for am in self.analysis_modules:
            result += am[key]

        return result

    def clear_tag_cache(self):

        for obj in self.all_objects():
            obj.clear_cache()

            for k in obj.behavior:
                obj.behavior[k].clear_cache()


    ################################################################################################
    #deprecated#####################################################################################
    ################################################################################################

    @deprecated_warning("add_behaviors_to_object function will be removed in coming versions. Please use obj.add_behaviors or obj.add_behavior instead.")
    def add_behaviors_to_object(self, b_dict, obj):
        obj.add_behaviors(b_dict)

    @deprecated_warning("add_behaviors_to_objects function will be removed in coming versions. Please use obj.add_behaviors or obj.add_behavior instead.")
    def add_behaviors_to_objects(self, b_dict, objs):
        for obj in objs:
            obj.add_behaviors(b_dict)

    '''
    def set_synapses_to_neuron_groups(self):# todo: move to synapse group __init__
        for ng in self.NeuronGroups:
            ng.afferent_synapses = {}
            ng.efferent_synapses = {}

            for sg in self.SynapseGroups:
                for tag in sg.tags+['All']:
                    ng.synapses(afferent, tag) = []
                    ng.efferent_synapses[tag] = []

            for sg in self.SynapseGroups:
                if sg.dst.BaseNeuronGroup == ng:
                    for tag in sg.tags+['All']:
                        ng.synapses(afferent, tag).append(sg)

                if sg.src.BaseNeuronGroup == ng:
                    for tag in sg.tags+['All']:
                        ng.efferent_synapses[tag].append(sg)
    '''


from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.StorageManager.StorageManager import *
from PymoNNto.Exploration.Evolution.TxtHighlighter import *
from PyQt5 import QtCore
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import pyqtgraph as pg
import numpy as np
from distutils.dir_util import copy_tree

def unique(l):
    return list(sorted(set(l)))


def get_words(blocks):
    return unique([word for word in ' '.join(blocks).split(' ') if word != ''])

#UI_base.Add_element(InteractiveScatter(...), sidebar, stretch)
class InteractiveScatter(pg.GraphicsLayoutWidget):#canvas object

    def __init__(self,default_x='index', default_y='score', coloration_param='list', *args, **kwargs):#
        super().__init__(*args, **kwargs)

        self.default_x = default_x
        self.default_y = default_y
        self.coloration_param = coloration_param
        self.global_coloration = True

        self.storage_manager_groups = []

        self.initialize_plot()

        self.scatter_clicked_event = None
        self.scatter_double_clicked_event = None

        self.selected_sml = []


    def scatter_clicked(self, plot, points):
        print('clicked')
        if len(points) > 0:
            clicked_x = points[-1]._data[0]
            clicked_y = points[-1]._data[1]

            self.clicked_sm = points[-1]._data[3]
            self.clicked_smg = points[-1]._data[4]

            x = []
            y = []
            self.selected_sml = [self.clicked_sm]

            if hasattr(self, 'gene_keys'):#search for other individuals with same genome
                clicked_id=self.clicked_sm.load_param('id')
                if clicked_id is not None:
                    data = self.clicked_smg.get_multi_param_list(['#SM#','id',self.default_x, self.default_y]+self.gene_keys)
                    indx=np.where(data[1]==clicked_id)[0]
                    current_genome = data[4:,indx]
                    for i in range(data.shape[1]):
                        if np.sum(data[4:, i]) == np.sum(current_genome):
                            x.append(data[2, i])
                            y.append(data[3, i])
                            self.selected_sml.append(data[0, i])


            x.append(clicked_x)
            y.append(clicked_y)
            c = [pg.mkBrush(0, 0, 255, 100) for _ in range(len(x))]
            c[-1] = pg.mkBrush(0, 0, 255, 255)
            self.scatter2.setData(x=x,y=y, brush=c)#[clicked_x], y=[clicked_y]  # set second scatter to new selection

            for i, d in enumerate(self.scatter2.data):
                d[3] = self.selected_sml[i] #storage manager

            self.scatter2.data[0][3] = self.clicked_sm
            if self.scatter_clicked_event is not None:
                self.scatter_clicked_event(self.clicked_sm)

            self.selected_sml=list(set(self.selected_sml))#remove dublicats

    def copy_selected(self, target_folder):
        # copy subdirectory example
        #from_directory = "/a/b/c"
        #to_directory = "/x/y/z"

        if len(target_folder)>0:
            for sm_src in self.selected_sml:
                print(sm_src.absolute_path, " ==> ")
                sm_dst = StorageManager(target_folder)
                copy_tree(sm_src.absolute_path, sm_dst.absolute_path)



    def copy_dialog(self):
        dlg = QDialog()
        dlg.setWindowTitle("Enter target folder name (Data/StorageManager/#target#/...)")
        layout = QVBoxLayout()

        input_le = QLineEdit("evo_copy")
        layout.addWidget(input_le)

        def btn_clicked():
            self.copy_selected(input_le.text())
            dlg.close()

        btn = QPushButton("copy data")
        btn.clicked.connect(btn_clicked)

        layout.addWidget(btn)
        dlg.setLayout(layout)
        dlg.resize(300, 50)
        dlg.exec()



    def scatter_double_clicked(self, plot, points):#scatter2
        if len(points) > 0:
            sm = points[-1]._data[3]

            print('double clicked', sm)

            #sm = self.clicked_sm

            txt = open(sm.absolute_path + sm.config_file_name, 'r').read()

            layout = QVBoxLayout()
            pte = QPlainTextEdit()
            #############################################################################################################################################################

            coloration = get_settings('txt_coloration')
            coloration_list = []
            #try:
            coloration_list = [[eval('QColor'+color), text] for color, text in coloration.items()]
            #coloration_list = [[QColor(0, 200, 0), words], [QColor(0, 0, 200), sentences]]
            pte.highlight = TxtHighlighter(pte.document(), coloration_list)
            #except:
            #    pass

            #sentences = ['fox eats meat.', 'boy drinks juice.', 'penguin likes ice.', 'man drives car.',
            #             'the fish swims.', 'plant loves rain.', 'parrots can fly.']
            #words = get_words(sentences)
            ###########################################################################################################################################
            pte.setPlainText(txt)
            pte.setReadOnly(True)
            layout.addWidget(pte, stretch=100)

            if len(self.selected_sml)>1:
                copy_btn = QPushButton('Copy selected ('+str(len(self.selected_sml))+')')
                layout.addWidget(copy_btn, stretch=1)
                copy_btn.clicked.connect(self.copy_dialog)

            dlg = QDialog()
            dlg.setLayout(layout)
            dlg.resize(1200, 800)
            dlg.exec()

            if self.scatter_double_clicked_event is not None:
                self.scatter_double_clicked_event(self.clicked_sm)

    def change_axis_param(self, axis_name, param):

        dx = self.default_x
        dy = self.default_y
        dc = self.coloration_param

        try:
            if axis_name == 'bottom':
                self.default_x = param
                self.plot.getAxis(axis_name).setLabel(text=param)

            if axis_name == 'left':
                self.default_y = param
                self.plot.getAxis(axis_name).setLabel(text=param)

            if axis_name == 'color':
                self.coloration_param = param

            self.refresh_data()

            return param
        except:
            self.default_x = dx
            self.default_y = dy
            self.coloration_param = dc



    def get_all_params(self):
        result_dict = {}
        for smg in self.storage_manager_groups:
            for param in smg.get_all_params():
                result_dict[param] = True
        return result_dict.keys()


    def axis_dialog(self, axis_name):
        dlg = QDialog()
        dlg.setWindowTitle('Select ' + axis_name + ' axis parameter')
        layout = QVBoxLayout()

        listwidget = QListWidget()

        if axis_name=='color':
            listwidget.addItems(['list'])

        listwidget.addItems(self.get_all_params())

        layout.addWidget(listwidget)

        def btn_clicked():
            self.change_axis_param(axis_name, listwidget.currentItem().text())
            dlg.close()

        btn = QPushButton('set new axis parameter')
        btn.clicked.connect(btn_clicked)

        layout.addWidget(btn)
        dlg.setLayout(layout)
        dlg.resize(300, 300)
        dlg.exec()


    def initialize_plot(self):
        self.setBackground((255, 255, 255))

        #if tooltip_message is not None:
        #    self.ci.setToolTip(tooltip_message)

        self.clicked_sm = None
        self.clicked_smg = ''

        self.plot = self.addPlot(row=0, col=0)

        # trendline
        self.mean_line = pg.PlotCurveItem([],pen=(0,0,0,255))
        self.plot.addItem(self.mean_line)

        # minmax area
        self.max_line = pg.PlotCurveItem([],pen=(0,0,0,255))
        self.plot.addItem(self.max_line)

        self.min_line = pg.PlotCurveItem([],pen=(0,0,0,255))
        self.plot.addItem(self.min_line)

        self.min_max_fill = pg.FillBetweenItem(curve1=self.max_line, curve2=self.min_line, brush=(0, 0, 0, 30))
        self.plot.addItem(self.min_max_fill)

        #on click highlight scatter item
        self.scatter2 = pg.ScatterPlotItem(size=10, brush=pg.mkBrush(0, 0, 255, 255))
        self.plot.addItem(self.scatter2)
        self.scatter2.sigClicked.connect(self.scatter_double_clicked)


        self.plot.getAxis('bottom').setLabel(text=self.default_x)
        self.plot.getAxis('left').setLabel(text=self.default_y)

        def bottom_axis_clicked(ev):
            self.axis_dialog('bottom')

        self.plot.getAxis('bottom').mouseClickEvent = bottom_axis_clicked

        def left_axis_clicked(ev):
            self.axis_dialog('left')

        self.plot.getAxis('left').mouseClickEvent = left_axis_clicked




    def find_same_path_smg(self, smg):
        found = None
        for attached_smg in self.storage_manager_groups:
            if smg.absolute_path == attached_smg.absolute_path:
                found = attached_smg

        return found

    def remove_StorageManagerGroup(self, smg):
        found = self.find_same_path_smg(smg)
        if found is not None:
            self.storage_manager_groups.remove(found)
            self.plot.removeItem(found.scatter)
            self.plot.removeItem(found.error_bar)
            self.update_indices()
            self.scatter2.clear()

    def get_smg(self, tag):
        for smg in self.storage_manager_groups:
            if smg.Tag == tag:
                return smg

    def add_StorageManagerGroup(self, smg):

        found = self.find_same_path_smg(smg)

        if found is None:
            smg.error_bar = pg.ErrorBarItem()
            smg.scatter = pg.ScatterPlotItem(size=10, name=smg.Tag)#, brush=pg.mkBrush(np.random.randint(0,255), np.random.randint(0,255), 255, 120)
            smg.scatter.sigClicked.connect(self.scatter_clicked)
            #self.plot.legend.addItem(smg.scatter, smg.Tag)
            self.plot.addItem(smg.scatter)
            self.plot.addItem(smg.error_bar)
        else:
            smg.error_bar = found.error_bar
            smg.scatter = found.scatter
            smg.color = found.color
            self.storage_manager_groups.remove(found)

        if not hasattr(smg, 'color'):
            smg.color = (np.random.randint(0, 255), np.random.randint(0, 255), np.random.randint(0, 255))

            #print('ck', smg.color[0], smg.color[1], smg.color[2], 120)
            #color = pg.mkBrush(smg.color[0], smg.color[1], smg.color[2], 120)
            #color = pg.mkBrush(np.random.randint(0, 255), np.random.randint(0, 255), 255, 120)
            #smg.scatter.setData(brush=color)

        self.storage_manager_groups.append(smg)

        self.update_indices()

        self.plot.removeItem(self.scatter2)#bring to front
        self.plot.addItem(self.scatter2)
        self.scatter2.clear()


    def update_indices(self):
        return
        #for i, smg in enumerate(self.storage_manager_groups):
        #    smg.add_virtual_multi_parameter('index', i)

    def add_mean_and_variance(self):
        if self.default_x=='index':
            result_lists = smg.get_multi_param_list([self.default_x, self.default_y], remove_None=True).astype(np.float64)

            mx, my = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.average(a)')
            vx, vy = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.var(a)')




    def add_trendline(self, smg):
        if self.default_x=='generation' or self.default_x=='gen':
            result_lists = smg.get_multi_param_list([self.default_x, self.default_y], remove_None=True).astype(np.float64)

            xa, ya = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.average(a)')
            xs, ymi = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.min(a)')
            xs, yma = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.max(a)')

            self.mean_line.setData(xa, ya)
            self.min_line.setData(xa, ymi)
            self.max_line.setData(xa, yma)

            self.mean_line.setVisible(True)
            self.min_line.setVisible(True)
            self.max_line.setVisible(True)
            self.min_max_fill.setVisible(True)
        else:
            self.mean_line.setVisible(False)
            self.min_line.setVisible(False)
            self.max_line.setVisible(False)
            self.min_max_fill.setVisible(False)

    def refresh_data(self):

        #refresh and get global min max color
        color_vals = []
        for smg in self.storage_manager_groups:
            if self.coloration_param != 'list' and self.global_coloration:
                color_vals += smg.get_param_list(self.coloration_param, remove_None=True)

            smg.refresh()

        #actual refresh
        for smg in self.storage_manager_groups:

            params=['#SM#', self.default_x, self.default_y]

            if self.coloration_param != 'list':
                params += [self.coloration_param]

            data = smg.get_multi_param_list(params, remove_None=True)

            if self.coloration_param == 'list':
                c = pg.mkBrush(smg.color[0], smg.color[1], smg.color[2], 120)
            else:
                if self.global_coloration:
                    min=np.min(color_vals)
                    max=np.max(color_vals)
                else:
                    min=np.min(data[3])
                    max=np.max(data[3])
                d=max-min
                if d==0:
                    d=1
                c=[pg.mkBrush(255-255/d*(c-min), 255/d*(c-min), 0, 120) for c in data[3]]

            smg.scatter.setData(x=data[1], y=data[2], brush=c)  #

            #error = pg.ErrorBarItem(x=x, y=y, top=top, bottom=bottom, beam=0.5)
            if self.default_x == 'index' and len(data[1])>0:
                meany= np.mean(data[2])
                stdy = np.std(data[2])
                smg.error_bar.setData(x=np.mean(data[1]), y=meany, top=stdy, bottom=stdy, beam=0.3)
                smg.error_bar.setVisible(True)
            else:
                smg.error_bar.setVisible(False)

            for i, d in enumerate(smg.scatter.data):  # set ids to each point (d[3] very ugly coding...) (each point is a set, not an object)
                d[3] = data[0][i] #storage manager
                d[4] = smg  #storage manager group

        self.scatter2.clear()

        single_smg = len(self.storage_manager_groups) == 1
        if single_smg:
            self.add_trendline(self.storage_manager_groups[0])






from PymoNNto.NetworkCore.Neuron_Group import *
from PymoNNto.NetworkCore.Behavior import *
import numpy as np

class Synaptic_Normalization(Behavior):

    def initialize(self, neurons):
        self.syn_type = self.parameter('syn_type', 'GLU', neurons)
        self.clip_max = self.parameter('clip_max', None, neurons)
        neurons.weight_norm_factor = neurons.vector()+self.parameter('norm_factor', 1.0, neurons)

    def iteration(self, neurons):
        normalize_synapse_attr('W', 'W', neurons.weight_norm_factor, neurons, self.syn_type)
        for s in neurons.synapses(afferent, self.syn_type):
            s.W = np.clip(s.W, 0, self.clip_max)






def normalize_synapse_attr(src_attr, target_attr, target_value, neurons, synapse_type):
    neurons.temp_weight_sum = neurons.vector()

    for s in neurons.synapses(afferent, synapse_type):
        s.dst.temp_weight_sum += np.sum(np.abs(getattr(s, src_attr)), axis=1)

    neurons.temp_weight_sum /= target_value

    for s in neurons.synapses(afferent, synapse_type):
        setattr(s, target_attr, getattr(s, target_attr) / (s.dst.temp_weight_sum[:, None] + (s.dst.temp_weight_sum[:, None] == 0)))




def normalize_synapse_attr_efferent(src_attr, target_attr, target_value, neurons, synapse_type):
    neurons.temp_weight_sum = neurons.vector()

    for s in neurons.synapses(efferent, synapse_type):
        s.src.temp_weight_sum += np.sum(np.abs(getattr(s, target_attr)), axis=0)

    neurons.temp_weight_sum /= target_value

    for s in neurons.neurons.synapses(efferent, synapse_type):
        setattr(s, src_attr, getattr(s, src_attr) / (s.src.temp_weight_sum + (s.src.temp_weight_sum == 0)))

'''
from PymoNNto import *

net = Network(tag='Network')

ng = NeuronGroup(tag='Neuron', net=net, behavior={}, size=10)
sg = SynapseGroup(tag='GLU', src=ng, dst=ng, net=net, behavior={})

net.initialize()

sg.W = sg.matrix('random')

print(np.sum(sg.W,0))
print(np.sum(sg.W,1))

normalize_synapse_attr_efferent('W','W', 1, ng, 'GLU')
normalize_synapse_attr('W','W', 1, ng, 'GLU')

print(np.sum(sg.W,0))
print(np.sum(sg.W,1))

import matplotlib.pyplot as plt

plt.matshow(sg.W)
plt.show()
'''


from PymoNNto import *
from PymoNNto.NetworkBehavior.Basics.BasicHomeostasis import *



class NOX_Diffusion(Instant_Homeostasis):

    def partition_sum(self, neurons):
        neurons._temp_act_sum = neurons.vector()
        for sg, sg_rf in zip(self.subgroups, self.subgroups_rf):
            sg._temp_act_sum += np.mean(sg_rf.output_new)
        return neurons._temp_act_sum

    def initialize(self, neurons):
        super().initialize(neurons)

        self.subgroups = self.split_grid_into_sub_group_blocks(steps=[5, 5, 1])

        receptive_field = 5
        self.subgroups_rf = []
        for sg in self.subgroups:
            sg_rf = neurons.get_subgroup_receptive_field_mask(sg, [receptive_field, receptive_field, receptive_field])
            self.subgroups_rf.append(sg_rf)

        neurons.nox = neurons.vector()
        self.adjustment_param = 'nox'

        self.measurement_param = self.parameter('mp', 'self.partition_sum(n)', None)

        self.set_threshold(self.parameter('th_nox', 0, neurons))
        self.adj_strength = -self.parameter('eta_nox', 0.002, neurons)

    def iteration(self, neurons):
        neurons.nox.fill(0)
        super().iteration(neurons)


from PymoNNto.Exploration.Network_UI.TabBase import *

class event_tab(TabBase):

    def __init__(self, title='Events'):
        super().__init__(title)


    def initialize(self, Network_UI):
        self.event_tab = Network_UI.add_tab(title=self.title)

        self.plot = Network_UI.tab.add_plot(title='Events', x_label='t (iterations)', y_label='events')

        self.current_pos = pg.PlotCurveItem([0, 0], [0, 0], pen=(100,100,100,255))
        self.plot.addItem(self.current_pos)

        self.plot.hideAxis('left')

        self.event_count = 0
        self.label_x_positions = []

        self.ypos = 2.0


    def update(self, Network_UI):
        if self.event_tab.isVisible():

            self.current_pos.setData([0, Network_UI.network.iteration], [0, 0])

            #update_plot
            for i,e in enumerate(Network_UI.event_list):
                if i>self.event_count:

                    if e.start_t!=e.end_t:
                        curve = pg.PlotCurveItem([e.start_t, e.end_t], [0, 0], fillLevel=10, brush=(30, 30, 30, 30), pen=(30, 30, 30, 30))
                        self.plot.addItem(curve)
                    else:

                        if '(False)' in e.tag:
                            line = pg.InfiniteLine(pos=e.start_t, angle=90, pen=(100, 0, 0, 100), span=(0, 0.1))
                        elif '(True)' in e.tag:
                            line = pg.InfiniteLine(pos=e.start_t, angle=90, pen=(0, 100, 0, 100), span=(0, 0.1))
                        else:
                            line = pg.InfiniteLine(pos=e.start_t, angle=90, pen=(0, 0, 100, 100), span=(0, 0.1))


                        self.plot.addItem(line)

                        text = pg.TextItem(text=e.tag.replace('(False)', '(OFF)').replace('(True)', '(ON)'), anchor=(0.5, 0))
                        text.setPos(e.start_t, self.ypos)
                        self.ypos += 0.5
                        if self.ypos > 10.0:
                            self.ypos = 2.0
                        self.plot.addItem(text)

                        self.label_x_positions.append(e.start_t)


                    self.event_count+=1





from PymoNNto.Exploration.Evolution.Devices.Evolution_Device_Multi_Thread import *
import sys
from PymoNNto.Exploration.Evolution.SSH_Functions import *

def ssh_thread_worker(slave_file, name, ssh_string, root_folder, conn):
    print('ssh thread started')
    while True:
        time.sleep(np.random.rand())
        if conn.poll():
            genome = conn.recv()
            try:

                genome_str=''
                for gene in genome:
                    genome_str+=' '+ str(gene) + '=' + str(genome[gene])

                ssh = get_ssh_connection(ssh_string)
                cmd = 'cd ' + name + '/' +root_folder+ '; '
                cmd += 'python ' + slave_file + genome_str#' genome=' + get_gene_id(genome)
                ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(cmd)
                results = get_response(ssh_stdout, ssh_stderr)
                ssh.close()

                success = False
                for line in results:
                    if 'evolution score set to #' in line:
                        parts = line.split('#')
                        if len(parts) == 3:
                            genome['score'] = float(parts[1])
                            success = True

                if success:
                    sm = StorageManager(main_folder_name=genome['evo_name'], folder_name=get_gene_file(genome), print_msg=False, add_new_when_exists=False)
                    sm.save_param_dict(genome)
                    conn.send([genome, 'success'])
                else:
                    conn.send([genome, 'no score found in ssh thread results'+str(results)])

            except:
                conn.send([genome, 'thread error'])

        conn.send([None, 'idle'])
        time.sleep(1.0)


class Evolution_Device_SSH(Evolution_Device_Multi_Thread):

    def __init__(self, device_string, parent):
        super().__init__(device_string, parent)

        temp=sys.argv[0].replace(get_data_folder().replace('Data', ''), '')
        temp='/'.join(temp.split('/')[:-1])#remove file name
        self.root_folder = temp #evolution_master_path_from_project_folder

    def initialize(self):
        self.parent_conn, child_conn = Pipe()
        self.process = Process(target=ssh_thread_worker, args=(self.parent.slave_file, self.parent.name, self.device_string, self.root_folder, child_conn))

    def initialize_device_group(self):
        transfer_project(self.parent.name, self.device_string)



from PymoNNto.Exploration.StorageManager.StorageManager import *

class StorageManager2:

    # %nr %rnd
    def __init__(self, folder, subfolder='folder+"_"+id', mainfolder='StorageManager', datafolder=get_data_folder(),
                 print_path=True):
        if os.path.exists(folder):
            path = folder
        else:
            self.path = self._ensure_folder(datafolder + '/')
            self.path = self._ensure_folder(self.path + mainfolder + '/')
            self.path = self._ensure_folder(self.path + folder + '/')

            subfolder = self._eval_subfolder(folder, subfolder)
            new = not os.path.exists(self.path + subfolder + '/')
            self.path = self._ensure_folder(self.path + subfolder + '/')

            if new:
                self.save('time', time.time())

        if print_path:
            print(self.path)

        self.config = ConfigParser()
        self.config.read(self.path + 'config.ini')


    def _eval_subfolder(self, folder, subfolder):
        try:
            if 'id' in subfolder:
                id = len(os.listdir(self.path))

            if 'rnd' in subfolder:
                rnd = str(int(np.random.rand() * 10000))

            subfolder = eval(subfolder)
        except:
            pass

        return subfolder


    @property
    def folder(self):
        return self.path


    @property
    def directory(self):
        return self.path


    def _ensure_folder(self, folder):
        if not os.path.exists(folder):
            try:
                os.mkdir(folder)
            except:
                raise Exception(folder + ' can not be created')
        return folder


    def _get_next_key_file(self, key):
        result = self.path + key + '.*'
        i = 0
        while len(glob.glob(result)) > 0:
            result = self.path + key + '_' + str(i) + '.*'
            i += 1
        return result


    # add network save pickle compile...
    def save(self, key, data, mainkey='', ):
        if type(data) == str or type(data) == int or type(data) == float or type(data) == bool:

            value = str(data)

            if not self.config.has_section(mainkey):
                self.config.add_section(mainkey)
            self.config.set(mainkey, key, str(value).replace('%', 'percent'))
            with open(self.absolute_path + self.config_file_name, 'w') as configfile:
                self.config.write(configfile)

        else:
            file = self._get_next_key_file(mainkey + key)

            if type(data) == np.ndarray:
                np.save(file + '.npy', arr=obj)

            elif type(data) == image:
                imageio.imwrite(file + '.png', image.astype(np.uint8))

            else:  # obj
                pickle.dump(data, open(file + '.obj', 'wb'))


    def load(self, key, default=None, mainkey=''):
        try:
            s = self.config.get(mainkey, key)
        except:
            return default
        if len(s) > 0:
            if s == 'True':
                return True
            if s == 'False':
                return False
            try:
                if '.' in s:
                    return float(s)
                else:
                    return int(s)
            except:
                return s
        return s


    # do not use! => render_video searches for key*.png (right order!!!) (creation date?)

    # self.get_next_frame_name(key)
    # def get_next_frame_key(self, videokey):
    #    if not key in self.frame_counter:
    #        self.frame_counter[key] = 0
    #    self.frame_counter[key] += 1
    #    return self.absolute_path+key+'{}'.format(self.frame_counter[key]-1)


    def render_video(self, key, delete_images, reset_frame_counter=True):
        call(
            ['ffmpeg', '-i', self.path + key + '_%d.png', '-c:v', 'libx264', '-strict', '-2', '-preset', 'slow', '-pix_fmt',
             'yuv420p', '-vf', 'scale=trunc(iw/2)*2:trunc(ih/2)*2', '-f', 'mp4', self.absolute_path + key + '.mp4'],
            shell=True)
        if delete_images:
            files = os.listdir(self.path)
            for f in files:
                if key in f and '.png' in f:
                    os.remove(self.absolute_path + f)


from PymoNNto.Exploration.Evolution.Devices.Evolution_Device import *
from multiprocessing import Process, Pipe
#import subprocess
#import time
#import os
import numpy as np
import sys
import traceback

def local_thread_worker(slave_file, evo_name, static_genome, conn):
    #try:
    #    import psutil as psu
    #    parent = psu.Process()
    #    parent.nice(psu.BELOW_NORMAL_PRIORITY_CLASS)
    #except:
    #    print('not able to set BELOW_NORMAL_PRIORITY_CLASS')
    print('local thread started')
    while True:
        time.sleep(np.random.rand()/10)#avoid creating same storage manager files at same time
        if conn.poll():
            evo_id, generation, genome = conn.recv()
            try:
                execute_local_file(slave_file, evo_name, evo_id, generation, genome, static_genome)
                conn.send([evo_id, 'success'])
            except Exception as e:
                error_type = str(sys.exc_info()[0])
                msg = str(sys.exc_info()[1])
                conn.send([genome, 'thread error '+error_type+' '+msg+'\r\n'+traceback.format_exc()])

        conn.send([None, 'idle'])


class Evolution_Device_Multi_Thread(Evolution_Device):

    def initialize(self):
        self.parent_conn, child_conn = Pipe()
        self.process = Process(target=local_thread_worker, args=(self.parent.slave_file, self.parent.name, self.parent.inactive_genome_info, child_conn))

    def main_loop_update(self):

        if self.parent_conn.poll():
            evo_id, thread_msg = self.parent_conn.recv()
            if thread_msg == 'success':
                self.score_processing(evo_id)
            elif thread_msg == 'idle':
                current_individual = self.parent.get_next_individual()
                if current_individual is not None:
                    arguments = [current_individual.id, self.parent.Breed_And_Select.generation, current_individual.genome]
                    self.parent_conn.send(arguments)
            else:
                self.parent.error_event(evo_id, thread_msg)

    def start(self):
        self.process.start()

    def stop(self):
        self.process.stop()#???





import sys
from PyQt5.QtCore import QRegExp
from PyQt5.QtWidgets import QApplication, QDialog, QPlainTextEdit, QVBoxLayout, QTextEdit
from PyQt5.QtGui import QColor, QTextCharFormat, QSyntaxHighlighter

class TxtHighlighter(QSyntaxHighlighter):

    def __init__(self, document, coloration_list={}, base_char_color=QColor(0, 0, 0)):#{['a'],QColor()} #, words=[], sentences=[]
        QSyntaxHighlighter.__init__(self, document)
        self.base_char_color=base_char_color
        self.rules = []

        for color, string_list in coloration_list:
            self.rules += [(QRegExp(r'%s' % s), color) for s in string_list]


    def highlightBlock(self, text):
        self.setFormat(0, len(text), self.base_char_color)
        for expression, color in self.rules:
            index = expression.indexIn(text, 0)
            while index >= 0:
                index = expression.pos(0)
                length = len(expression.cap(0))
                _format = QTextCharFormat()
                _format.setForeground(color)#_format.setBackground(color)
                self.setFormat(index, length, _format)
                index = expression.indexIn(text, index + length)
        self.setCurrentBlockState(0)


import sys
import os

__evolution_genome__ = {}
__evo_name__ = None
__evo_id__ = None
__evo_generation__ = None
__get_gene_mode__ = False

def is_evolution_mode():
    return not gene('__evo_name__', None) == None

def get_gene_mode():
    return __get_gene_mode__

def get_evo_name():
    if __evo_name__ is None:
        update_evolution_parameters()
    return __evo_name__

def get_evo_id():
    if __evo_id__ is None:
        update_evolution_parameters()
    return __evo_id__

def get_evo_generation():
    if __evo_generation__ is None:
        update_evolution_parameters()
    return __evo_generation__

def get_genome():
    if len(__evolution_genome__)==0:
        update_evolution_parameters()
    return __evolution_genome__

def __print_gene(key, default):
    print('#get_gene#' + str(key) + '#' + str(default) + '#' + str(type(default)) + '#')

def get_gene(key, default):
    if len(__evolution_genome__)==0:
        update_evolution_parameters()

    if not key in __evolution_genome__:
        __evolution_genome__[key] = default

    if get_gene_mode():
        __print_gene(key, default)

    return cast_to_default(__evolution_genome__[key], default)


def cast_to_default(str_value, default):
    if isinstance(default, float) or isinstance(default, int):
        return float(str_value)
    else:
        return str_value

def gene(key, default):
    return get_gene(key, default)

def set_genome(genome):
    update_evolution_parameters()
    global __evolution_genome__
    __evolution_genome__ = genome
    if get_gene_mode():
        print('#set_genome#')
        for key, default in genome.items():
            __print_gene(key, default)
        print('#genome_set#')

def execute_local_file(file, evo_name, evo_id, evo_generation=0, genome={}, static_genome={}, get_gene_mode=False, get_output=False):
    cmd = 'python3 '+file

    cmd+=' __evo_name__' + '="' + evo_name+'"'
    cmd+=' __evo_id__' + '=' + str(evo_id)
    cmd+=' __evo_generation__' + '=' + str(evo_generation)

    if get_gene_mode:
        cmd += ' __get_gene_mode__'

    for key, value in genome.items():
        cmd+=' ' + str(key) + '="' + str(value)+'"'

    for key, value in static_genome.items():
        cmd+=' ' + str(key) + '="' + str(value)+'"'

    print(os.getcwd(), file)

    if get_output:
        return os.popen(cmd).read()
    else:
        os.system(cmd)#python3 file.py __evo_name__=myname __evo_id__=10001 gene1=5 gene2=a gene3=...

def update_evolution_parameters():# pythonfile.py __evo_name__=myname __evo_id__=10001 gene1=5 gene2=a gene3=...
    global __evo_name__
    global __evo_id__
    global __evo_generation__
    global __evolution_genome__
    global __get_gene_mode__

    for arg in sys.argv[1:]:
        if arg == '__get_gene_mode__':
            __get_gene_mode__ = True
        if '=' in arg:
            key_value = arg.split('=')
            if len(key_value)==2:
                key = key_value[0]
                value = key_value[1]
                if key == '__evo_name__':
                    __evo_name__ = value
                elif key == '__evo_id__':
                    __evo_id__ = value
                elif key == '__evo_generation__':
                    __evo_generation__ = value
                else:
                    __evolution_genome__[key] = value


def get_gene_file(evo_name, id=None):
    file = evo_name +' id' + str(id)
    return file



def set_score(score, info={}, sm=None):
    update_evolution_parameters()

    global __evo_name__
    global __evo_id__
    global __evo_generation__
    global __evolution_genome__

    if sm is None and __evo_name__ is not None and __evo_id__ is not None:
        import PymoNNto.Exploration.StorageManager.StorageManager as storage_manager
        sm = storage_manager.StorageManager(main_folder_name=__evo_name__, folder_name=get_gene_file(__evo_name__, __evo_id__), print_msg=False, add_new_when_exists=False)

    if sm is not None:
        sm.save_param('score', str(score))
        sm.save_param('evo_name', str(__evo_name__))
        sm.save_param('id', str(__evo_id__))
        sm.save_param('generation', str(__evo_generation__))
        sm.save_param_dict(__evolution_genome__)
        sm.save_param_dict(info)

    print('score set to:', score, info)

    __evo_name__ = None
    __evo_id__ = None
    __evo_generation__ = None
    __evolution_genome__ = {}


'''
def set_score(score, non_evo_storage_manager=None, _genome=None, info={}):
    global evolution_genome

    if _genome is not None:
        evolution_genome = _genome

    if 'evo_name' in evolution_genome:#evolution_genome is not None
        if 'score' in evolution_genome:#only when setscore is called multiple times on accident...
            evolution_genome.pop('score')
        #if 'evo_name' in evolution_genome:# and 'generation' in evolution_genome and 'id' in evolution_genome
        sm = StorageManager(main_folder_name=get_gene('evo_name', None), folder_name=get_gene_file(evolution_genome), print_msg=False, add_new_when_exists=False)
        evolution_genome['score'] = score
        sm.save_param_dict(evolution_genome)
        sm.save_param_dict(info)
        print('evolution score set to #'+str(score)+'#')
        #else:
        #    print('cannot save score', str(score), 'to file: "evo_name", "gen" or "id" not in genome')
    else:
        print('score='+str(score)+' (no genome found)')
        if non_evo_storage_manager is not None:
            non_evo_storage_manager.save_param('score', score)
            non_evo_storage_manager.save_param_dict(info)

    # reset and reload when next get_gene is called

    evolution_genome = None
'''



###############################old

#def get_default_genome():
#    return default_genome

#def get_gene_file(gene):
#    file = gene['evo_name']
#    if 'generation' in  gene:
#        file+=' gen' + str(gene['generation'])
#    if 'id' in gene:
#        file += ' id' + str(gene['id'])
#    return file

#def load_genome2():
#    for arg in sys.argv:
#        if 'genome=' in arg:
#            set_evolution_genome_from_string(arg[7:])

#def set_evolution_genome_from_string(gene_str):
#    #print('gene_str', gene_str)
#    gene_dict = {}
#    partial_str = gene_str.split('#')
#    for key_value_str in partial_str:
#        kv_partial_str = key_value_str.split('@')
#        if len(kv_partial_str) == 2:
#            if kv_partial_str[0] != 'score':
#                gene_dict[kv_partial_str[0]] = kv_partial_str[1] #keys and values are strings!!! have to be casted to default type by get_gene!
#
#    set_genome(gene_dict)

#######################################

#def get_gene_id(gene):
#    id = ''
#    for key, value in gene.items():
#        id += '#'+key+'@'+str(value)
#    return id+'#'


#def execute_local_file(file, evo_name, evo_id, evo_generation, genome, static_genome={}): #file.py __evo_name__=myname __evo_id__=10001 gene1=5 gene2=a gene3=...
#    py_file = open(file, "r")
#    file_content = py_file.read()
#    py_file.close()

    #remove old infomation
#    for arg in sys.argv:
#        if '=' in arg:
#            sys.argv.remove(arg)

#    global __evo_name__
#    global __evo_id__
#    global __evo_generation__
#    global __evolution_genome__
#    __evo_name__ = None
#    __evo_id__ = None
#    __evo_generation__ = None
#    __evolution_genome__= {}

    #add new information
#    sys.argv.append('__evo_name__' + '=' + evo_name)
#    sys.argv.append('__evo_id__' + '=' + str(evo_id))
#    sys.argv.append('__evo_generation__' + '=' + str(evo_generation))

#    for key, value in genome.items():
#        sys.argv.append(str(key) + '=' + str(value))

#    for key, value in static_genome.items():
#        sys.argv.append(str(key) + '=' + str(value))

#    exec(file_content)


from PymoNNto.Exploration.Network_UI.Network_UI import *
from PymoNNto.Exploration.Network_UI.TabBase import *




class UI_sidebar_network_selector():

    def __init__(self, add_color_dict={'output': (255, 255, 255)}):
        if type(add_color_dict) is str:
            add_color_dict={add_color_dict:(255,255,255)}
        self.add_color_dict = add_color_dict

        self.compiled = {}
        for param in self.add_color_dict:
            self.compiled[param] = compile('n.'+param, '<string>', 'eval')

    #def add_recorder_variables(self, neuron_group, Network_UI):
    #    for module in self.sub_modules:
    #        module.add_recorder_variables(neuron_group, Network_UI)

    #def on_selected_neuron_changed(self, NetworkUI):
    #    for module in self.sub_modules:
    #        module.on_selected_neuron_changed(NetworkUI)

    def get_clicked_neuron_id(self, Network_UI, event):
        group = event.currentItem.neuron_group
        w = group.width
        h = group.height
        d = group.depth
        y_temp = int(
            (h * d - 1) - np.trunc(event.pos().y()))

        neuron_select_x = np.clip(int(np.trunc(event.pos().x())), 0, w - 1)
        neuron_select_y = np.clip(int(y_temp - np.trunc(y_temp / h)), 0, h - 1)
        neuron_select_z = np.clip(int(np.trunc(y_temp / h)), 0, d - 1)

        h_abs = np.clip(int(y_temp), 0, h * d - 1)
        return (h_abs) * w + neuron_select_x


    def addRow(self, Network_UI, layout, ng):
        row = QHBoxLayout()
        layout.addLayout(row, stretch=1)

        ng.visible_cb = QCheckBox()
        ng.uilabel = QLabel(ng.tag)
        ng.color_select_box = Analytics_Results_Select_ComboBox(ng, 'classifier', first_entry='group color')
        ng.slider = QSlider(1)

        ng.slider.setMinimum(0)
        ng.slider.setMaximum(100)
        ng.slider.setMaximumWidth(100)
        ng.slider.setSliderPosition(100)
        ng.slider.mouseReleaseEvent = Network_UI.static_update_func
        ng.slider.setToolTip('scale neuron-group plots up and down (only visualization)')

        row.addWidget(ng.visible_cb, stretch=1)
        row.addWidget(ng.uilabel, stretch=10)
        row.addWidget(ng.color_select_box, stretch=10)
        row.addWidget(ng.slider, stretch=10)

        ng.visible_cb.setChecked(True)
        ng.visible_cb.group = ng

        def visible_cb_changed(event):
            ng.visible_cb.group.is_visible = ng.visible_cb.isChecked()

        ng.visible_cb.stateChanged.connect(visible_cb_changed)

    def initialize(self, Network_UI):

        self.config_dlg = QDialog()
        self.config_dlg.setWindowTitle('config')
        layout = QVBoxLayout()
        self.config_dlg.setLayout(layout)
        self.config_dlg.resize(600, 100)

        #self.config_dlg
        #dlg.exec()

        self.selected_label = Network_UI.sidebar.add_widget(QLabel(), stretch=1)
        link_font = QFont()
        link_font.setUnderline(True)
        self.selected_label.setFont(link_font)
        self.selected_label.setStyleSheet("color: rgb(0,0,200)")

        def show_dlg(event):
            self.config_dlg.exec()
        self.selected_label.mouseReleaseEvent = show_dlg


        for ng in Network_UI.network.NeuronGroups:
            #layout = Network_UI.sidebar.add_row()

            self.addRow(Network_UI, layout, ng)

            # Network_UI.sidebar.add_widget(, stretch=0)

            #Network_UI.sidebar.add_widget(ng.color_select_box, stretch=0)

            #self.add_cb(Network_UI, ng)

            #ng.color_select_box.hide()
            #ng.uilabel.hide()
            #ng.hide_cb.hide()

            #layout.removeWidget(ng.color_select_box)
            #layout.removeWidget(ng.uilabel)

            #ng.color_select_box.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
            #ng.uilabel.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)

            #ng.color_select_box.adjustSize()
            #ng.uilabel.adjustSize()
            #layout.adjustSize()

            #Network_UI.sidebar.set_parent_layout()

        self.plt = Network_UI.sidebar.add_plot(tooltip_message='white: active neurons\r\ncolor: neuron classification or base color\r\ngreen: selected neuron', stretch=100)



        #selected...
        self.sel_rect = QGraphicsRectItem()#

        self.plt.addItem(self.sel_rect)
        self.sel_rect.setPen(QPen(QColor(0, 255, 0), 3))


        #self.sel_rect.setParentItem()

        y = 0
        x = 0
        for i, ng in enumerate(Network_UI.network.NeuronGroups):

            ng.image_item = self.plt.add_image()
            ng.image_item.neuron_group=ng
            ng.classification = ng.vector()
            image = np.array(ng['Neuron_Classification_Colorizer', 0].get_color_list(ng.classification, format='[r,g,b]'))
            image = np.reshape(image, (ng.height * ng.depth, ng.width, 3))
            ng.image_item.setImage(np.rot90(image, 3), levels=(0, 255))

            ng.image_item.setPos(x, y)
            y += ng.height + 5

            def mce(event):
                self.last_id = self.get_clicked_neuron_id(Network_UI, event)
                self.last_group = event.currentItem.neuron_group#Network_UI.network[event.currentItem.neuron_group_tag, 0]
                Network_UI.select_neuron(self.last_group, self.last_id)
                #for neurons in Network_UI.network.NeuronGroups:
                #    neurons.color_select_box.hide()
                #    neurons.uilabel.hide()
                #    neurons.hide_cb.hide()
                #self.last_group.color_select_box.show()
                #self.last_group.uilabel.show()
                #self.last_group.hide_cb.show()


            def mdce(event):
                if self.last_group.classification is not None:
                    selected_class = self.last_group.classification[self.last_id]
                    Network_UI.select_neuron_class(self.last_group, selected_class)

            ng.image_item.mouseClickEvent = mce
            ng.image_item.mouseDoubleClickEvent = mdce



            #ng.image_item.setRect(0, -i * 10, ng.width, ng.height)

        #self.image_item.neuron_group = Network_UI.network.exc_neurons1

        #if Network_UI.group_display_count is not None:
        #    self.group_display_count=Network_UI.group_display_count

        #self.sub_modules = []
        #for i in range(self.group_display_count):
        #    self.sub_modules.append(sidebar_neuron_grid_submodule(self.add_color_dict))
        #    self.sub_modules[-1].initialize(Network_UI, np.minimum(i, len(Network_UI.group_tags)))

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def on_selected_neuron_changed(self, NetworkUI):
        return

    def update(self, Network_UI):

        for i, ng in enumerate(Network_UI.network.NeuronGroups):
            n = ng  # for eval operations

            ng.classification = ng.color_select_box.get_selected_result()
            if ng.classification is None:
                ng.classification = ng.vector()
            image = np.array(ng['Neuron_Classification_Colorizer', 0].get_color_list(ng.classification, format='[r,g,b]'))

            # selected
            if Network_UI.selected_neuron_group() == ng:
                self.selected_label.setText('Selected: ' + ng.tag)
                self.sel_rect.setRect(ng.image_item.boundingRect())
                self.sel_rect.setPos(ng.image_item.pos())
                # self.sel_rect.setPen(QColor=QColor(0,255,0,255))
                for i in range(3):
                    image[Network_UI.selected_neuron_mask(), i] = Network_UI.neuron_select_color[i]

            # parameter
            for param, color in self.add_color_dict.items():
                try:
                    data = eval(self.compiled[param])

                    if (type(data) == np.ndarray and data.dtype == np.dtype('bool')) or (type(data) == bool):
                        for i in range(3):
                            image[data, i] += color[i]
                    else:
                        for i in range(3):
                            image[:, i] += color[i] * data
                except:
                    pass

            image = np.reshape(image, (ng.height * ng.depth, ng.width, 3))  # ng.depth
            ng.image_item.setImage(np.rot90(image, 3), levels=(0, 255))
            #ng.image_item.setPos(0, i * 1000)
            #ng.image_item.setRect(0, -i * 10, ng.width, ng.height)

'''

class sidebar_neuron_grid_submodule(TabBase):

    def __init__(self, add_color_dict):
        self.add_color_dict = add_color_dict

        self.compiled = {}
        for param in self.add_color_dict:
            self.compiled[param] = compile('n.'+param, '<string>', 'eval')

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def on_selected_neuron_changed(self, Network_UI):
        tag = self.image_item.neuron_group_tag
        self.color_select_box.change_main_object(Network_UI.network[tag, 0])

    def get_clicked_neuron_id(self, Network_UI, event):
        group_tag = event.currentItem.neuron_group_tag
        w = Network_UI.network[group_tag, 0].width
        h = Network_UI.network[group_tag, 0].height
        d = Network_UI.network[group_tag, 0].depth
        y_temp = int(
            (h * d - 1) - np.trunc(event.pos().y()))

        neuron_select_x = np.clip(int(np.trunc(event.pos().x())), 0, w - 1)
        neuron_select_y = np.clip(int(y_temp - np.trunc(y_temp / h)), 0, h - 1)
        neuron_select_z = np.clip(int(np.trunc(y_temp / h)), 0, d - 1)

        h_abs = np.clip(int(y_temp), 0, h * d - 1)
        return (h_abs) * w + neuron_select_x

    def initialize(self, Network_UI, index):
        #just store for mcde event bug ('QGraphicsSceneMouseEvent' object has no attribute 'currentItem')
        self.last_id = 0
        self.last_group = None

        def mce(event):
            self.last_id = self.get_clicked_neuron_id(Network_UI, event)
            self.last_group = Network_UI.network[event.currentItem.neuron_group_tag, 0]
            Network_UI.select_neuron(self.last_group, self.last_id)

        def mdce(event):
            if self.last_group.classification is not None:
                selected_class = self.last_group.classification[self.last_id]
                Network_UI.select_neuron_class(self.last_group, selected_class)

        group_select_box = QComboBox()
        self.color_select_box = Analytics_Results_Select_ComboBox(Network_UI.network[Network_UI.group_tags[index], 0] ,'classifier', first_entry='group color')

        #Network_UI.sidebar.add_row()
        #Network_UI.sidebar.add_widget(group_select_box, stretch=0)
        #Network_UI.sidebar.add_widget(self.color_select_box, stretch=0)
        #Network_UI.sidebar.set_parent_layout()

        Network_UI.sidebar.add_row()
        Network_UI.sidebar.add_widget([group_select_box, self.color_select_box])
        Network_UI.sidebar.set_parent_layout()

        self.image_item = Network_UI.sidebar.add_plot(tooltip_message='white: active neurons\r\ncolor: neuron classification or base color\r\ngreen: selected neuron', stretch=100).add_image()

        self.image_item.neuron_group_tag = Network_UI.group_tags[index]
        Network_UI.neuron_visible_groups.append(Network_UI.group_tags[index])
        self.image_item.mouseClickEvent = mce

        self.image_item.mouseDoubleClickEvent = mdce


        def group_changed(select_index):
            tag = Network_UI.group_tags[select_index]
            Network_UI.select_neuron(Network_UI.network[tag, 0], 0)
            self.color_select_box.change_main_object(Network_UI.network[tag, 0])
            self.image_item.neuron_group_tag = tag
            Network_UI.get_visible_neuron_groups()[index] = tag

        group_select_box.addItems(Network_UI.group_tags)
        group_select_box.setCurrentIndex(index)
        group_select_box.currentIndexChanged.connect(group_changed)



    def update(self, Network_UI):
        group_tag = self.image_item.neuron_group_tag
        if len(Network_UI.network[group_tag]) > 0:

            group = Network_UI.network[group_tag, 0]
            n = group #for eval operations

            group.classification = self.color_select_box.get_selected_result()
            if group.classification is None:
                group.classification = group.vector()
            image = np.array(group['Neuron_Classification_Colorizer', 0].get_color_list(group.classification, format='[r,g,b]'))

            #selected
            if Network_UI.selected_neuron_group().tags[0] == group_tag:
                for i in range(3):
                    image[Network_UI.selected_neuron_mask(), i] = Network_UI.neuron_select_color[i]

            #parameter
            for param, color in self.add_color_dict.items():
                try:
                    data=eval(self.compiled[param])

                    if (type(data) == np.ndarray and data.dtype == np.dtype('bool')) or (type(data) == bool):
                        for i in range(3):
                            image[data, i] += color[i]
                    else:
                        for i in range(3):
                            image[:, i] += color[i]*data
                except:
                    pass



            image=np.reshape(image, (group.height*group.depth, group.width, 3))#group.depth
            self.image_item.setImage(np.rot90(image, 3), levels=(0, 255))

class UI_sidebar_neuron_grid_module():

    def __init__(self, group_display_count=1, add_color_dict={'output': (255, 255, 255)}):
        self.group_display_count = group_display_count
        if type(add_color_dict) is str:
            add_color_dict={add_color_dict:(255,255,255)}
        self.add_color_dict = add_color_dict

    def add_recorder_variables(self, neuron_group, Network_UI):
        for module in self.sub_modules:
            module.add_recorder_variables(neuron_group, Network_UI)

    def on_selected_neuron_changed(self, NetworkUI):
        for module in self.sub_modules:
            module.on_selected_neuron_changed(NetworkUI)

    def initialize(self, Network_UI):

        if Network_UI.group_display_count is not None:
            self.group_display_count=Network_UI.group_display_count

        self.sub_modules = []
        for i in range(self.group_display_count):
            self.sub_modules.append(sidebar_neuron_grid_submodule(self.add_color_dict))
            self.sub_modules[-1].initialize(Network_UI, np.minimum(i, len(Network_UI.group_tags)))

    def update(self, Network_UI):
        for module in self.sub_modules:
            module.update(Network_UI)

'''


import random
import numpy as np
from PymoNNto.Exploration.Evolution.Interface_Functions import *
from PymoNNto.Exploration.Evolution.Evolution_Individual import *

class Default_Breed_And_Select:

    def __init__(self, parent, death_rate, mutation, individual_count, constraints, additional_evo_params={}):
        self.parent = parent
        self.death_rate = death_rate
        self.mutation = mutation
        self.individual_count = individual_count
        self.generation = 1
        self.constraints = constraints

        if 'no_high_risk_inds' in additional_evo_params:
            self.avoid_high_risk_high_reward_individuals = additional_evo_params['no_high_risk_inds']
        else:
            self.avoid_high_risk_high_reward_individuals = True

        if 're_eval_inds' in additional_evo_params:
            self.re_evaluate_genomes = additional_evo_params['re_eval_inds']
        else:
            self.re_evaluate_genomes = True


    def natural_selection(self, individuals):
        result = individuals.copy()

        #sort (biggest score first)
        for i in range(len(result)-1):
            for j in range(len(result)-i-1):
                if result[i].score < result[i+1+j].score:
                    temp = result[i]
                    result[i] = result[i+1+j]
                    result[i+1+j] = temp

        kill_count = int(len(result) * self.death_rate)
        for _ in range(kill_count):
            result.pop(-1)#remove last (smallest score)

        return result

    def generate_new_generation(self, old_individuals):
        survivours = self.natural_selection(old_individuals)
        new_population = self.breed(survivours)

        self.generation += 1
        for individual in new_population:
            individual.id = self.parent.get_individual_id()

        return new_population


    def update_population(self):
        #print('update...', len(self.parent.non_scored_individuals), len(self.parent.running_individuals), len(self.parent.scored_individuals))
        if len(self.parent.non_scored_individuals) == 0 and len(self.parent.running_individuals) == 0:

            new_population = self.generate_new_generation(self.parent.scored_individuals)

            print('new generation:', self.generation)

            self.parent.non_scored_individuals = new_population
            self.parent.running_individuals = []
            self.parent.scored_individuals = []

            #if not self.re_evaluate_genomes:
            #    for s in survivours:
            #        found = self.parent.find(self.parent.non_scored_individuals, s)
            #        self.parent.non_scored_individuals.remove(found)
            #        self.parent.running_individuals.append(found)

            #        processing_genome = self.parent.add_name_gen_id_inactive_to_genome(s)
            #        set_score(s.score, _genome=processing_genome)#create files #s['score']
            #        self.parent.new_score_event(processing_genome)#move from non_scored_individuals to scored_individuals



    def get_offspring(self, ind1, ind2):
        child = Evolution_individual(genome=None)

        while not child.is_valid_genome(self.constraints):
            child.genome = {}

            for key in ind1.genome.keys():
                if key in ind2.genome.keys():
                    g1 = np.random.normal(ind1.genome[key], np.clip(ind1.genome[key] * self.mutation, 0, None))
                    g2 = np.random.normal(ind2.genome[key], np.clip(ind2.genome[key] * self.mutation, 0, None))
                    child.genome[key] = random.choice([g1, g2])
                else:
                    print('not the same genes', ind1, ind2)

        return child


    def breed(self, individuals):

        #weights are used to avoid high risk, high reward individuals.
        #only individuals with persistent performances over generations have high mate_chance and produce offspring
        weights = np.ones(len(individuals))
        if self.avoid_high_risk_high_reward_individuals:
            for i, ind in enumerate(individuals):
                if ind.mate_chance is None or ind.score is None:#score is none during initialization
                    ind.mate_chance = 1
                else:
                    ind.mate_chance = (ind.mate_chance + ind.score) / 2
                weights[i] = ind.mate_chance
        weights = weights / np.sum(weights)  # todo divide by zero for negative scores!


        result = []

        for ind in individuals:
            result.append(ind.get_copy())# only genome and mate_chances are copied

        new_count = self.individual_count - len(result)

        for i in range(new_count):
            parent1 = np.random.choice(individuals, 1, p=weights)[0]
            parent2 = np.random.choice(individuals, 1, p=weights)[0]
            result.append(self.get_offspring(parent1, parent2))

        return result




from PymoNNto.NetworkCore.Behavior import *
import copy

class Recorder(Behavior):
    initialize_last = True

    visualization_module_outputs = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.variables = {}


    def initialize(self, parent_obj):
        self.add_tag('recorder')

        variables = self.parameter('variables', None)
        if variables is None:
            variables = self.parameter('arg_0', None)
            if variables is None:
                variables = []

        if type(variables) is str:
            variables = [variables]

        self.gapwidth = self.parameter('gapwidth', 0)
        self.save_as_numpy = self.parameter('save_as_numpy', False)
        self.max_length = self.parameter('max_length', None)
        self.counter = 0
        self.new_data_available=False
        self.compiled = {}

        self.add_variables(variables)
        self.reset()

    def iteration(self, parent_obj):

        if parent_obj.recording:

            self.counter += 1
            if self.counter >= self.gapwidth:
                self.new_data_available = True
                self.counter = 0
                for v in self.variables:
                    if self.compiled[v] is None:
                        annotated_v = self.annotate_var_str(v, parent_obj)
                        self.compiled[v] = compile(annotated_v, '<string>', 'eval')
                    data = self.get_data_v(v, parent_obj)
                    if data is not None:
                        self.save_data_v(data, v)

        if self.max_length is not None:
            self.cut_length(self.max_length)

    def eq_split(self, eq, splitter):
        str = eq.replace(' ', '')
        parts = []
        str_buf = ''
        for s in str:
            if s in splitter:
                parts.append(str_buf)
                parts.append(s)
                str_buf = ''
            else:
                str_buf += s

        parts.append(str_buf)
        return parts

    def annotate_var_str(self, variable, parent_obj):
        splitter = ['*', '/', '+', '-', '%', ':', ';', '=', '!', '(', ')', '[', ']', '{', '}']
        annotated_var = ''
        for part in self.eq_split(variable, splitter):
            if hasattr(parent_obj, part):
                part = 'n.'+part
            annotated_var += part
        return annotated_var

    def add_varable(self, v):
        if self.save_as_numpy:
            self.variables[v] = np.array([])
        else:
            self.variables[v] = []
        self.compiled[v] = None

    def add_variables(self, vars):
        for v in vars:
            self.add_varable(v)

    def find_objects(self, key):
        result = []
        if key in self.variables:
            result.append(self.variables[key])
        return result

    def reset(self):
        for v in self.variables:
            self.add_varable(v)



    def is_new_data_available(self):
        if self.new_data_available:
            self.new_data_available = False
            return True
        else:
            return False

    def get_data_v(self, variable, parent_obj):
        n = parent_obj  # used for eval string "n.x"
        s = parent_obj
        neurons = parent_obj
        synapse = parent_obj
        try:
            return copy.copy(eval(self.compiled[variable]))
        except:
            return None

    def save_data_v(self, data, variable):
        if self.save_as_numpy:
            if len(self.variables[variable]) == 0:
                self.variables[variable] = np.array([data])
            else:
                self.variables[variable] = np.concatenate([self.variables[variable], [data]], axis=0)
        else:
            self.variables[variable].append(data)



    def cut_length(self, max_length):
        if max_length is not None:
            for v in self.variables:
                while len(self.variables[v]) > max_length:
                    if self.save_as_numpy:
                        print('not implemented')
                    else:
                        self.variables[v].pop(0)

    def swaped(self, name):
        return self.swap(self.variables[name])#getattr(self, name)

    def swap(self, x):
        return np.swapaxes(np.array(x), 1, 0)

    def clear(self):
        self.clear_recorder()

    def clear_recorder(self):
        print('clear')

        for v in self.variables:
            self.variables[v].clear()


class EventRecorder(Recorder): #10: EventRecorder(tag='my_event_recorder', variables=['spike']) #plt.plot(My_Network['spike.t', 0], My_Network['spike.i', 0], '.k')

    def find_objects(self, key):

        result = []
        if key in self.variables:
            result.append(self.variables[key])

        if type(key) is str and key[-2:] == '.t' and key[:-2] in self.variables:#time
            result.append(self.variables[key[:-2]][:, 0])

        if type(key) is str and key[-2:] == '.i' and key[:-2] in self.variables:#neuron index
            result.append(self.variables[key[:-2]][:, 1])

        return result

    def get_data_v(self, variable, parent_obj):
        n = parent_obj  # used for eval string "n.x"
        s = parent_obj
        neurons = parent_obj
        synapse = parent_obj

        data = eval(self.compiled[variable])

        indices = np.where(data != 0)
        iteration = np.ones_like(indices[0]) * parent_obj.iteration

        return np.stack((iteration, *indices), axis=1)

    def save_data_v(self, data, variable):
        if len(self.variables[variable]) == 0:
            self.variables[variable] = np.array(data)
        else:
            self.variables[variable] = np.concatenate([self.variables[variable], data], axis=0)




import numpy as np
from PymoNNto.NetworkCore.Network import *
from PymoNNto.NetworkCore.Behavior import *
from PymoNNto.NetworkCore.Neuron_Group import *
from PymoNNto.NetworkCore.Synapse_Group import *
from PymoNNto.NetworkCore.Analysis_Module import *

from PymoNNto.Exploration.HelperFunctions import *
from PymoNNto.Exploration.StorageManager.StorageManager import *

from PymoNNto.Exploration.Evolution.Evolution import *

import os
import shutil

from PymoNNto.Exploration.Network_UI import *

import pytest

#from PyQt5.QtCore import QLibraryInfo
#os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(
#    QLibraryInfo.PluginsPath
#)


folder = get_data_folder()+'/StorageManager/'
def clear_folder(f):
    if os.path.isdir(folder+f+'/'):
        shutil.rmtree(folder+f+'/')


class Counter(Behavior):
    def initialize(self, neurons):
        self.inc = self.parameter('inc', 1)
        neurons.count = neurons.vector()
    def iteration(self, neurons):
        neurons.count += self.inc

def get_sample_network():
    My_Network = Network()
    My_Neurons = NeuronGroup(net=My_Network, tag='my_neurons', size=100, behavior={
        1: Counter(inc='[2#I]'),
        2: Recorder('count')
    })
    My_Synapses = SynapseGroup(net=My_Network, src=My_Neurons, dst=My_Neurons, tag='GLUTAMATE')
    sm = StorageManager('test', random_nr=False, print_msg=False)
    My_Network.initialize(storage_manager=sm)

    return My_Network, My_Neurons, My_Synapses, sm


def test_basics():
    net = Network()
    NeuronGroup(net=net, tag='neurons', size=100, behavior={})
    NeuronGroup(net=net, tag='neurons2', size=50, behavior={})
    SynapseGroup(net=net, tag='syn', src='neurons', dst='neurons2', behavior={})

    net.initialize()

    zero_vec = net.neurons.vector()

    one_vec = net.neurons.vector(1)
    m10_vec = net.neurons.vector(-10)

    bool_vec = net.neurons.vector(bool)
    int_vec = net.neurons.vector(int)

    uni_vec = net.neurons.vector('uniform')#, plot=True
    rnd_vec = net.neurons.vector('random')#, plot=True
    log_vec = net.neurons.vector('lognormal(mean=1, sigma=1)')#, plot=True
    norm_vec = net.neurons.vector('normal(loc=1,scale=1)')#, plot=True

    assert len(zero_vec) == 100 and zero_vec[0] == 0
    assert len(one_vec) == 100 and one_vec[0] == 1
    assert len(m10_vec) == 100 and m10_vec[0] == -10

    assert bool_vec[0] == False and bool_vec.dtype == bool
    assert int_vec[0] == 0 and int_vec.dtype == int

    assert len(uni_vec) == 100 and uni_vec[0]>=0 and uni_vec[0]<=1
    assert len(rnd_vec) == 100 and rnd_vec[0]>=0 and rnd_vec[0]<=1
    assert len(log_vec) == 100
    assert len(norm_vec) == 100


    zero_mat = net.syn.matrix()

    one_mat = net.syn.matrix(1)
    m10_mat = net.syn.matrix(-10)

    bool_mat = net.syn.matrix(bool)
    int_mat = net.syn.matrix(int)

    uni_mat = net.syn.matrix('uniform')#, plot=True
    rnd_mat = net.syn.matrix('random')#, plot=True
    log_mat = net.syn.matrix('lognormal(mean=1, sigma=1)')#, plot=True
    norm_mat = net.syn.matrix('normal(loc=1,scale=1)')#, plot=True

    assert zero_mat.shape[0] == 50 and zero_mat.shape[1] == 100 and zero_mat[0, 0] == 0
    assert one_mat.shape[0] == 50 and one_mat.shape[1] == 100 and one_mat[0, 0] == 1
    assert m10_mat.shape[0] == 50 and m10_mat.shape[1] == 100 and m10_mat[0, 0] == -10

    assert bool_mat[0, 0] == False and bool_mat.dtype == bool
    assert int_mat[0, 0] == 0 and int_mat.dtype == int

    assert uni_mat.shape[0] == 50 and uni_mat.shape[1] == 100 and uni_mat[0, 0]>=0 and uni_mat[0, 0]<=1
    assert rnd_mat.shape[0] == 50 and rnd_mat.shape[1] == 100 and rnd_mat[0, 0]>=0 and rnd_mat[0, 0]<=1
    assert log_mat.shape[0] == 50 and log_mat.shape[1] == 100
    assert norm_mat.shape[0] == 50 and norm_mat.shape[1] == 100

def test_behavior_and_tagging():
    print()

    # basic network
    set_genome({'I': 1})
    My_Network, My_Neurons, My_Synapses, sm = get_sample_network()

    My_Network.simulate_iterations(1000)

    My_Network.deactivate_behaviors('Counter')
    My_Network.simulate_iterations(10)

    My_Network.activate_behaviors('Counter')
    My_Network.simulate_iterations(20)

    My_Network.recording_off()
    My_Network.simulate_iterations(30)

    assert My_Network.iteration == 1000+10+20+30
    assert np.mean(My_Neurons.count) == 1000+20+30

    assert My_Synapses.src == My_Neurons
    assert My_Synapses.dst == My_Neurons

    assert My_Neurons.synapses(afferent, 'GLUTAMATE') == [My_Synapses]

    assert len(My_Network.all_objects()) == 3

    #tagging system
    assert My_Network['my_neurons'] == [My_Neurons]
    assert len(My_Network['count', 0]) == My_Network.iteration-30

    My_Network.clear_recorder()
    assert len(My_Neurons['count', 0]) == 0

    assert My_Network.tag_shortcuts['my_neurons'] == My_Network['my_neurons']

    if os.path.isdir(folder+'test/'):
        shutil.rmtree(folder+'test/')



def test_storage_manager():
    print()

    clear_folder('test')

    sm = StorageManager('test', print_msg=False)
    assert os.path.isfile('Data/StorageManager/test/test/config.ini')

    sm.save_param('k', 0)
    assert sm.load_param('k') == 0

    sm.save_param_dict({'k1': 1, 'k2': 2, 'k3': 3})
    assert sm.load_param('k1') == 1
    assert sm.load_param('k2') == 2
    assert sm.load_param('k3') == 3

    sm2 = StorageManager('test', print_msg=False)
    sm2.save_param('k', 101)
    sm2.save_param('k1', 102)

    smg = StorageManagerGroup('test')

    smg.sort_by('k')

    k_pl = smg.get_param_list('k')
    assert k_pl == [0, 101]

    m_pl = smg.get_multi_param_list(['#SM#','k','k1'])
    assert m_pl[0, 0].load_param('k') == 0
    assert m_pl[0, 1].load_param('k') == 101

    assert m_pl[1, 0] == 0
    assert m_pl[1, 1] == 101

    assert m_pl[2, 0] == 1
    assert m_pl[2, 1] == 102

    clear_folder('test')


def test_add_remove_behaviors():
    print()

    set_genome({'I': 1})
    My_Network, My_Neurons, My_Synapses, sm = get_sample_network()

    My_Neurons.count *= 0
    My_Neurons.remove_behavior('Counter')
    My_Network.simulate_iterations(10)
    assert np.mean(My_Neurons.count) == 0

    My_Neurons.add_behavior(0.5, Counter(inc='2'), initialize=True)
    My_Network.simulate_iterations(10)
    assert np.mean(My_Neurons.count) == 2*10

    My_Neurons.remove_behavior('Counter')
    My_Neurons.add_behavior(4, Counter(inc='2'), initialize=False)
    assert np.mean(My_Neurons.count) == 2*10

    if os.path.isdir(folder+'test/'):
        shutil.rmtree(folder+'test/')

##########################UI

#pip install pytest-qt

'''
@pytest.fixture
def app(qtbot):
    My_Network, My_Neurons, My_Synapses, sm = get_sample_network()
    My_Network.simulate_iteration()

    ui_app = Network_UI(My_Network, modules=get_default_UI_modules(), title='test', storage_manager=sm, group_display_count=1, reduced_layout=False)
    ui_app.pause = True

    qtbot.addWidget(ui_app.main_window)

    return ui_app


def test_label(app):

    assert app.width == 1200

    def visible():
        return True

    for i in range(app.tabs.count()):
        app.tabs.widget(i).isVisible=visible

    for _ in range(10):
        for module in app.modules:
            module.update(app)

    if os.path.isdir(folder+'test/'):
        shutil.rmtree(folder+'test/')


############# Evolution


def test_evolution():
    print()

    #clear storage manager
    clear_folder('pytest_evo')

    genome = {'a': 1, 'b': 2, 'c': 2, 'd': 2, 'e':3}

    evo = Evolution(name='pytest_evo',
                    slave_file='Exploration/Evolution/example_slave.py',
                    individual_count=2,
                    mutation=0.04,
                    death_rate=0.5,
                    constraints=['b>=a', 'a<=1', 'b>= 2'],
                    inactive_genome_info={'info': 'my_info'},
                    start_genomes=[genome],
                    devices={'single_thread': 1,
                             #'multi_thread': 4,
                             #'ssh user@host.de': 0,
                             }
                    )


    for device in evo.devices:
        device.start()

    #run 10 full generations
    for _ in range(3):
        for device in evo.devices:
            device.main_loop_update()

    first_run_score = evo.scored_individuals[0].score

    #continue
    evo.load_state()

    #run to full generations
    for _ in range(3):
        for device in evo.devices:
            device.main_loop_update()

    second_run_score = evo.scored_individuals[0].score

    clear_folder('pytest_evo')

    start_score = 7
    assert first_run_score >= start_score
    assert second_run_score >= first_run_score
    assert evo.Breed_And_Select.generation == 3
    
'''